"""
This test file tests the lib.tokens.u2ftoken
This depends on lib.tokenclass
"""
from .base import MyTestCase
from privacyidea.lib.tokens.u2ftoken import U2fTokenClass
from privacyidea.lib.tokens.u2f import (check_registration_data,
                                        parse_registration_data, url_decode,
                                        check_response, parse_response_data)
from privacyidea.lib.token import init_token, remove_token, check_user_pass
from privacyidea.lib.user import User
from privacyidea.lib.policy import set_policy, SCOPE, ACTION, delete_policy
from privacyidea.lib.config import set_privacyidea_config
from privacyidea.lib.challenge import get_challenges
from privacyidea.lib.utils import hexlify_and_unicode, to_bytes
from privacyidea.lib.error import TokenAdminError
from privacyidea.lib import _
import binascii
from hashlib import sha256
from OpenSSL import crypto
import base64


REG_DATA = "BQRFnd8XtfZzsTK68VPK64Bcjiog_ZzyYNuzjaaGwpPnSpifxaqQV4_4IMxVlGS3CLoQmNAR41MSMxZHG0dENLRmQGnk4OqRxGRHmUOOLmDkGgdIJycQe79JCERV1gqGnWAOFBg_bH4WFSxZwnX-IMRcl3zW_X442QNrrdFySvXrba4wggIcMIIBBqADAgECAgQ4Zt91MAsGCSqGSIb3DQEBCzAuMSwwKgYDVQQDEyNZdWJpY28gVTJGIFJvb3QgQ0EgU2VyaWFsIDQ1NzIwMDYzMTAgFw0xNDA4MDEwMDAwMDBaGA8yMDUwMDkwNDAwMDAwMFowKzEpMCcGA1UEAwwgWXViaWNvIFUyRiBFRSBTZXJpYWwgMTM4MzExNjc4NjEwWTATBgcqhkjOPQIBBggqhkjOPQMBBwNCAAQ3jfx0DHOblHJO09Ujubh2gQZWwT3ob6-uzzjZD1XiyAob_gsw3FOzXefQRblty48r-U-o4LkDFjx_btwuSHtxoxIwEDAOBgorBgEEAYLECgEBBAAwCwYJKoZIhvcNAQELA4IBAQIaR2TKAInPkq24f6hIU45yzD79uzR5KUMEe4IWqTm69METVio0W2FHWXlpeUe85nGqanwGeW7U67G4_WAnGbcd6zz2QumNsdlmb_AebbdPRa95Z8BG1ub_S04JoxQYNLaa8WRlzN7POgqAnAqkmnsZQ_W9Tj2uO9zP3mpxOkkmnqz7P5zt4Lp5xrv7p15hGOIPD5V-ph7tUmiCJsq0LfeRA36X7aXi32Ap0rt_wyfnRef59YYr7SmwaMuXKjbIZSLesscZZTMzXd-uuLb6DbUCasqEVBkGGqTRfAcOmPov1nHUrNDCkOR0obR4PsJG4PiamIfApNeoXGYpGbok6nucMEYCIQC_yerJqB3mnuAJGfbdKuOIx-Flxr-VSQ2nAkUUE_50dQIhAJE2NL1Xs2oVEG4bFzEM86TfS7nkHxad89aYmUrII49V"
CLIENT_DATA_HASH = "eyJ0eXAiOiJuYXZpZ2F0b3IuaWQuZmluaXNoRW5yb2xsbWVudCIsImNoYWxsZW5nZSI6ImdXbndtYnFSMl9YOE91RFhId0dyQWNmUTBUajN4YTVfZ2RJMjBYcVlsdTg9Iiwib3JpZ2luIjoiaHR0cDovL2xvY2FsaG9zdDo1MDAwIiwiY2lkX3B1YmtleSI6IiJ9"
KEY_HANDLE_HEX = "61655467367048455a45655a51343475594f51614230676e4a78423776306b4952465857436f61645941345547443973666859564c466e436466346778467958664e6239666a6a5a4132757430584a4b396574747267"
APP_ID = "http://localhost:5000"


class U2FHelperFuncTestCase(MyTestCase):

    def test_01_check_reg_data(self):
        attestation_cert = "3082013c3081e4a003020102020a4790128000115595735230" \
                           "0a06082a8648ce3d0403023017311530130603550403130c" \
                           "476e756262792050696c6f74301e170d313230383134313" \
                           "8323933325a170d3133303831343138323933325a303131" \
                           "2f302d0603550403132650696c6f74476e756262792d302" \
                           "e342e312d34373930313238303030313135353935373335" \
                           "323059301306072a8648ce3d020106082a8648ce3d030107" \
                           "034200048d617e65c9508e64bcc5673ac82a6799da3c1446" \
                           "682c258c463fffdf58dfd2fa3e6c378b53d795c4a4dffb4" \
                           "199edd7862f23abaf0203b4b8911ba0569994e101300a06" \
                           "082a8648ce3d0403020347003044022060cdb6061e9c2226" \
                           "2d1aac1d96d8c70829b2366531dda268832cb836bcd30dfa" \
                           "0220631b1459f09e6330055722c8d89b7f48883b9089b88d" \
                           "60d1d9795902b30410df"

        cdata_str = """{"typ": "navigator.id.finishEnrollment",
                        "challenge": "vqrS6WXDe1JUs5_c3i4-LkKIHRr-3XVb3azuA5TifHo",
                        "cid_pubkey": {
                            "kty": "EC",
                            "crv": "P-256",
                            "x": "HzQwlfXX7Q4S5MtCCnZUNBw3RMzPO9tOyWjBqRl4tJ8",
                            "y": "XVguGFLIZx1fXg3wNqfdbn75hi4-_7-BxhMljw42Ht4"},
                        "origin": "http://example.com"}"""
        my_app_id = 'http://example.com'

        app_id_hash = hexlify_and_unicode(sha256(to_bytes(my_app_id)).digest())
        self.assertEqual(app_id_hash,
                         "f0e6a6a97042a4f1f1c87f5f7d44315b2d852c2df5c7991cc66241bf7072d1c4")
        attestation_cert = crypto.load_certificate(crypto.FILETYPE_ASN1,
                                                   binascii.unhexlify(attestation_cert))
        client_data_str = ''.join(cdata_str.split())
        client_data_hash = hexlify_and_unicode(sha256(to_bytes(client_data_str)).digest())
        self.assertEqual(client_data_hash,
                         "4142d21c00d94ffb9d504ada8f99b721f4b191ae4e37ca0140f696b6983cfacb")

        user_pub_key = "04b174bc49c7ca254b70d2e5c207cee9cf174820ebd77ea3c65508c26da51b657c1cc6b952f8621697936482da0a6d3d3826a59095daf6cd7c03e2e60385d2f6d9"
        key_handle = "2a552dfdb7477ed65fd84133f86196010b2215b57da75d315b7b9e8fe2e3925a6019551bab61d16591659cbaf00b4950f7abfe6660e2e006f76868b772d70c25"
        signature = "304502201471899bcc3987e62e8202c9b39c33c19033f7340352dba80fcab017db9230e402210082677d673d891933ade6f617e5dbde2e247e70423fd5ad7804a6d3d3961ef871"

        r = check_registration_data(attestation_cert, my_app_id, client_data_str,
                                    user_pub_key, key_handle, signature)
        self.assertEqual(r, True)

    def test_02_parse_reg_date(self):
        attestation_cert, user_pub_key, key_handle, signature, description \
            = parse_registration_data(REG_DATA)
        self.assertEqual(description, 'Yubico U2F EE Serial 13831167861')
        self.assertEqual(signature,
                     '3046022100bfc9eac9a81de69ee00919f6dd2ae388c7e165c6bf95490da702451413fe7475022100913634bd57b36a15106e1b17310cf3a4df4bb9e41f169df3d698994ac8238f55')
        self.assertEqual(key_handle, '69e4e0ea91c4644799438e2e60e41a07482727107bbf49084455d60a869d600e14183f6c7e16152c59c275fe20c45c977cd6fd7e38d9036badd1724af5eb6dae')
        self.assertEqual(user_pub_key, '04459ddf17b5f673b132baf153caeb805c8e2a20fd9cf260dbb38da686c293e74a989fc5aa90578ff820cc559464b708ba1098d011e353123316471b474434b466')

    def test_03_url_decode(self):
        dec = url_decode("SGFsbG8sIGRhcyBpc3QgZWluIFRlc3QuLi4")
        self.assertEqual(dec, b"Hallo, das ist ein Test...")

    def test_04_check_response(self):
        # According to
        # https://fidoalliance.org/specs/fido-u2f-v1.0-nfc-bt-amendment
        # -20150514/fido-u2f-raw-message-formats.html#authentication-example
        app_id = "https://gstatic.com/securitykey/a/example.com"
        user_pub_key = \
            "04d368f1b665bade3c33a20f1e429c7750d5033660c019119d29aa4ba7abc04aa7c80a46bbe11ca8cb5674d74f31f8a903f6bad105fb6ab74aefef4db8b0025e1d"
        cdata_str = """{"typ": "navigator.id.getAssertion",
                             "challenge": "opsXqUifDriAAmWclinfbS0e-USY0CgyJHe_Otd7z8o",
                             "cid_pubkey": {
                                 "kty": "EC",
                                 "crv": "P-256",
                                 "x": "HzQwlfXX7Q4S5MtCCnZUNBw3RMzPO9tOyWjBqRl4tJ8",
                                 "y": "XVguGFLIZx1fXg3wNqfdbn75hi4-_7-BxhMljw42Ht4"},
                             "origin": "http://example.com"}"""
        client_data_str = ''.join(cdata_str.split())
        counter = 1
        user_presence_byte = b'\x01'
        signature = "304402204b5f0cd17534cedd8c34ee09570ef542a353df4436030ce43d406de870b847780220267bb998fac9b7266eb60e7cb0b5eabdfd5ba9614f53c7b22272ec10047a923f"
        r = check_response(user_pub_key, app_id, client_data_str, signature,
                           counter, user_presence_byte)
        self.assertEqual(r, True)

    def test_05_parse_response(self):
        response = "0100000001304402204b5f0cd17534cedd8c34ee09570ef542a353df4436030ce43d406de870b847780220267bb998fac9b7266eb60e7cb0b5eabdfd5ba9614f53c7b22272ec10047a923f"
        up, count, sig = parse_response_data(response)
        self.assertTrue(up, "\x01")
        self.assertTrue(count, 1)
        self.assertTrue(sig, response[5:])


class U2FTokenTestCase(MyTestCase):

    def test_00_users(self):
        self.setUp_user_realms()

        set_privacyidea_config("u2f.appId", APP_ID)

    def test_01_create_token(self):
        pin = "test"
        # Init step1
        token = init_token({"type": "u2f",
                            "pin": pin})
        serial = token.token.serial

        self.assertEqual(token.type, "u2f")

        prefix = U2fTokenClass.get_class_prefix()
        self.assertEqual(prefix, "U2F")

        info = U2fTokenClass.get_class_info()
        self.assertEqual(info.get("type"), "u2f")

        info = U2fTokenClass.get_class_info("type")
        self.assertEqual(info, "u2f")

        idetail = token.get_init_detail()
        detail_serial = idetail.get("serial")
        self.assertEqual(serial, detail_serial)

        registerRequest = idetail.get("u2fRegisterRequest")
        version = registerRequest.get("version")
        self.assertEqual(version, "U2F_V2")
        challenge = registerRequest.get("challenge")
        self.assertEqual(len(challenge), 43)

        # Init step 2
        token = init_token({"type": "u2f",
                            "serial": serial,
                            "regdata": REG_DATA,
                            "clientdata": CLIENT_DATA_HASH})
        idetail = token.get_init_detail()
        subject = idetail.get("u2fRegisterResponse").get("subject")
        self.assertEqual(subject, 'Yubico U2F EE Serial 13831167861')

        # check the tokeninfo of the attestation certificate
        issuer = token.get_tokeninfo("attestation_issuer")
        subject = token.get_tokeninfo("attestation_subject")
        serial = token.get_tokeninfo("attestation_serial")
        self.assertEqual(issuer, "CN=Yubico U2F Root CA Serial 457200631")
        self.assertEqual(subject, "CN=Yubico U2F EE Serial 13831167861")
        self.assertEqual(serial, "946265973")

        #
        # Do some authentication
        #

        # challenge
        # check challenge
        is_chalrequest = token.is_challenge_request(pin)
        self.assertEqual(is_chalrequest, True)

        # create challenge
        res, message, t_id, reply_dict = token.create_challenge()
        attributes = reply_dict.get("attributes")
        self.assertTrue(res)
        expected_text = _("Please confirm with your U2F token ({0!s})").format("Yubico U2F EE Serial 13831167861")
        self.assertEqual(message, expected_text)
        self.assertEqual(len(t_id), 20)
        u2f_sign_request = attributes.get("u2fSignRequest")
        version = u2f_sign_request.get("version")
        self.assertEqual(version, "U2F_V2")
        key_handle = u2f_sign_request.get("keyHandle")
        key_handle_hex = hexlify_and_unicode(key_handle)
        self.assertTrue("appId" in u2f_sign_request, u2f_sign_request)
        self.assertTrue("challenge" in u2f_sign_request, u2f_sign_request)
        self.assertTrue("keyHandle" in u2f_sign_request, u2f_sign_request)
        self.assertEqual(key_handle_hex, KEY_HANDLE_HEX)
        self.assertEqual(u2f_sign_request.get("appId"), APP_ID)
        self.assertEqual(len(u2f_sign_request.get("challenge")), 43)

    def test_02_parse_regdata(self):
        client_data = "eyJ0eXAiOiJuYXZpZ2F0b3IuaWQuZmluaXNoRW5yb2xsbWVudCIsImNoYWxsZW5nZSI6IlNna3pUekdyYnNVREUyNEJSMV9kUTRYbXJtNTVqU2MzVml3Sm5DRjVmWm8iLCJvcmlnaW4iOiJodHRwczovL2RlbW8ueXViaWNvLmNvbSIsImNpZF9wdWJrZXkiOiJ1bnVzZWQifQ"
        reg_data = "BQT3NET2RTTcgzAiZRW5gkg3TT6mgQBepZl96iMtj-nXU25VdwBXCL1EjWOY-q1M76vT_iX9ebDhkZ1kvosbi3_AQGVopI2hcyIsc8q-KpzerJIZgWtN25bCy6g_hTk_M1khCjQGaiGJFwnk8GIn2OnkNOJRe7V00Q9PBZHn5mFwfFwwggJEMIIBLqADAgECAgRVYr6gMAsGCSqGSIb3DQEBCzAuMSwwKgYDVQQDEyNZdWJpY28gVTJGIFJvb3QgQ0EgU2VyaWFsIDQ1NzIwMDYzMTAgFw0xNDA4MDEwMDAwMDBaGA8yMDUwMDkwNDAwMDAwMFowKjEoMCYGA1UEAwwfWXViaWNvIFUyRiBFRSBTZXJpYWwgMTQzMjUzNDY4ODBZMBMGByqGSM49AgEGCCqGSM49AwEHA0IABEszH3c9gUS5mVy-RYVRfhdYOqR2I2lcvoWsSCyAGfLJuUZ64EWw5m8TGy6jJDyR_aYC4xjz_F2NKnq65yvRQwmjOzA5MCIGCSsGAQQBgsQKAgQVMS4zLjYuMS40LjEuNDE0ODIuMS41MBMGCysGAQQBguUcAgEBBAQDAgUgMAsGCSqGSIb3DQEBCwOCAQEArBbZs262s6m3bXWUs09Z9Pc-28n96yk162tFHKv0HSXT5xYU10cmBMpypXjjI-23YARoXwXn0bm-BdtulED6xc_JMqbK-uhSmXcu2wJ4ICA81BQdPutvaizpnjlXgDJjq6uNbsSAp98IStLLp7fW13yUw-vAsWb5YFfK9f46Yx6iakM3YqNvvs9M9EUJYl_VrxBJqnyLx2iaZlnpr13o8NcsKIJRdMUOBqt_ageQg3ttsyq_3LyoNcu7CQ7x8NmeCGm_6eVnZMQjDmwFdymwEN4OxfnM5MkcKCYhjqgIGruWkVHsFnJa8qjZXneVvKoiepuUQyDEJ2GcqvhU2YKY1zBFAiEAqqVKbLnZuWYyzjcsb1YnHEyuk-dmM77Q66iExrj8h2cCIHAvpisjLj-D2KvnZZcIQ_fFjFj9OX5jkfmJ65QVQ9bE"
        cert, user_pub_key, key_handle, signature, description = \
            parse_registration_data(reg_data)
        self.assertEqual(user_pub_key,
                         '04f73444f64534dc8330226515b98248374d3ea681005ea5997dea232d8fe9d7536e5577005708bd448d6398faad4cefabd3fe25fd79b0e1919d64be8b1b8b7fc0')
        self.assertEqual(key_handle,
                         '6568a48da173222c73cabe2a9cdeac9219816b4ddb96c2cba83f85393f3359210a34066a21891709e4f06227d8e9e434e2517bb574d10f4f0591e7e661707c5c')
        self.assertEqual(description, 'Yubico U2F EE Serial 1432534688')

        client_data_str = base64.b64decode(client_data+"==")
        r = check_registration_data(cert, "https://demo.yubico.com",
                                    client_data_str,
                                    user_pub_key, key_handle, signature)
        self.assertTrue(r)

        # modify signature
        broken_sig = 'ff' + signature[2:]
        with self.assertRaisesRegex(Exception,
                                    'Error checking the signature of the registration data.'):
            check_registration_data(cert, "https://demo.yubico.com", client_data_str,
                                    user_pub_key, key_handle, broken_sig)

    def test_03_missing_app_id(self):
        set_privacyidea_config("u2f.appId", '')
        pin = "test"
        token = init_token({"type": "u2f",
                            "pin": pin})
        with self.assertRaises(TokenAdminError):
            _detail = token.get_init_detail()

    def test_04_broken_reg_data(self):
        reg_data = 'BgT3NET2RTTcgzAiZRW5gkg3TT6mgQBepZl96iMtj-nXU25VdwBXCL1EjW' \
                   'OY-q1M76vT_iX9ebDhkZ1kvosbi3_AQGVopI2hcyIsc8q-KpzerJIZgWtN' \
                   '25bCy6g_hTk_M1khCjQGaiGJFwnk8GIn2OnkNOJRe7V00Q9PBZHn5mFwfF' \
                   'wwggJEMIIBLqADAgECAgRVYr6gMAsGCSqGSIb3DQEBCzAuMSwwKgYDVQQD' \
                   'EyNZdWJpY28gVTJGIFJvb3QgQ0EgU2VyaWFsIDQ1NzIwMDYzMTAgFw0xND' \
                   'A4MDEwMDAwMDBaGA8yMDUwMDkwNDAwMDAwMFowKjEoMCYGA1UEAwwfWXVi' \
                   'aWNvIFUyRiBFRSBTZXJpYWwgMTQzMjUzNDY4ODBZMBMGByqGSM49AgEGCC' \
                   'qGSM49AwEHA0IABEszH3c9gUS5mVy-RYVRfhdYOqR2I2lcvoWsSCyAGfLJ' \
                   'uUZ64EWw5m8TGy6jJDyR_aYC4xjz_F2NKnq65yvRQwmjOzA5MCIGCSsGAQ' \
                   'QBgsQKAgQVMS4zLjYuMS40LjEuNDE0ODIuMS41MBMGCysGAQQBguUcAgEB' \
                   'BAQDAgUgMAsGCSqGSIb3DQEBCwOCAQEArBbZs262s6m3bXWUs09Z9Pc-28' \
                   'n96yk162tFHKv0HSXT5xYU10cmBMpypXjjI-23YARoXwXn0bm-BdtulED6' \
                   'xc_JMqbK-uhSmXcu2wJ4ICA81BQdPutvaizpnjlXgDJjq6uNbsSAp98ISt' \
                   'LLp7fW13yUw-vAsWb5YFfK9f46Yx6iakM3YqNvvs9M9EUJYl_VrxBJqnyL' \
                   'x2iaZlnpr13o8NcsKIJRdMUOBqt_ageQg3ttsyq_3LyoNcu7CQ7x8NmeCG' \
                   'm_6eVnZMQjDmwFdymwEN4OxfnM5MkcKCYhjqgIGruWkVHsFnJa8qjZXneV' \
                   'vKoiepuUQyDEJ2GcqvhU2YKY1zBFAiEAqqVKbLnZuWYyzjcsb1YnHEyuk-' \
                   'dmM77Q66iExrj8h2cCIHAvpisjLj-D2KvnZZcIQ_fFjFj9OX5jkfmJ65QVQ9bE'
        with self.assertRaisesRegex(Exception,
                                    'The registration data is in a wrong format.'):
            parse_registration_data(reg_data)


class MultipleU2FTokenTestCase(MyTestCase):
    app_id = 'https://localhost:5000'
    reg_data1 = ("BQSUSmNE4buL5xIqMlaRyJSMHhSvr37LpTT2e-zxyVoY21dzl1gZwcTws9I8r"
                 "CCKwHV-j9dt3NsWIPypid8PsKWsQMU-IQlPNvVsMXxXyhMLjBoUvdiDZShlo5"
                 "w-P6_i_bT2mDF9O07q2lMW-AXGyIKKSrlsh-1oJBklRUYQ0sbfyEgwggK8MII"
                 "BpKADAgECAgQDrfASMA0GCSqGSIb3DQEBCwUAMC4xLDAqBgNVBAMTI1l1Ymlj"
                 "byBVMkYgUm9vdCBDQSBTZXJpYWwgNDU3MjAwNjMxMCAXDTE0MDgwMTAwMDAwM"
                 "FoYDzIwNTAwOTA0MDAwMDAwWjBtMQswCQYDVQQGEwJTRTESMBAGA1UECgwJWX"
                 "ViaWNvIEFCMSIwIAYDVQQLDBlBdXRoZW50aWNhdG9yIEF0dGVzdGF0aW9uMSY"
                 "wJAYDVQQDDB1ZdWJpY28gVTJGIEVFIFNlcmlhbCA2MTczMDgzNDBZMBMGByqG"
                 "SM49AgEGCCqGSM49AwEHA0IABBmeh5wWLbfcOe5KQqBGFqWzCf7KCS92vglI-"
                 "W1ulcrkzGXNVKBZz73HybMbKx1sGER5wsBh9BiqlUtZaiwc-hejbDBqMCIGCS"
                 "sGAQQBgsQKAgQVMS4zLjYuMS40LjEuNDE0ODIuMS43MBMGCysGAQQBguUcAgE"
                 "BBAQDAgQwMCEGCysGAQQBguUcAQEEBBIEEPormdyeOUJXj5JKMNI8QRgwDAYD"
                 "VR0TAQH_BAIwADANBgkqhkiG9w0BAQsFAAOCAQEAKOuzZ_7R2PDiievKn_bYB"
                 "1fGDprlfLFyjJscOMq7vYTZI32oMawhlJ8PLfwMMWv9sXWzbmOiK7tYDq3KUo"
                 "DQeYQOWh4lcmJaO_uHYDPb-yKpack4uJzhcTWUAKElLZcCqRKT1UUZ6WDdIs6"
                 "KJ-sF6355t1DAAv7ZAWtxHsmtdFAb2RTLvo7ZVxKBt09E6wd85h7LBquFqXJV"
                 "Jn7o45gr9D8Msho4LSNeueTObbKYxAVCUEAjKyth4QzXDGIVvAO36UBxtw4S0"
                 "cR_lmVaLvmdTOVafxtLH_kU7hNtnmEgRxSIZGmIgEQxFmU4ibhkhtnJyf-8k4"
                 "VFNWmzRXRLjKC0NzBEAiBcVPalM6D9t0JFcDFEJW4OiMr945SDuiJFaxlwi9M"
                 "SawIgdpy0pATSpLXaiFfUqxoFXaMfo-oSqssdZdcWIlWGCpI")
    client_data1 = ('eyJjaGFsbGVuZ2UiOiJJeF9HbWo1SkVFYktDR3k0QV9BVVZFYWRGTWNJam'
                    'hBOC1meTJha1ZFaUFRIiwib3JpZ2luIjoiaHR0cHM6Ly9sb2NhbGhvc3Q6'
                    'NTAwMCIsInR5cCI6Im5hdmlnYXRvci5pZC5maW5pc2hFbnJvbGxtZW50In0')

    reg_data2 = ('BQREKCB4dlbzcqkU7T1talLdlX8pV-aJa0ijSpM2JIXg490mvhXVRquJuSkhr'
                 'iz0YFrBsB0FjtbTejlsHeKO5de2QKfcG86b738Bq_RntwwH9Nf1VrsFHrdHnh'
                 '57-9-z7dTGKVD57W_LQch2QtjgZcPslUq19kbz-PXs-WJxgRNhMMUwggK8MII'
                 'BpKADAgECAgQDrfASMA0GCSqGSIb3DQEBCwUAMC4xLDAqBgNVBAMTI1l1Ymlj'
                 'byBVMkYgUm9vdCBDQSBTZXJpYWwgNDU3MjAwNjMxMCAXDTE0MDgwMTAwMDAwM'
                 'FoYDzIwNTAwOTA0MDAwMDAwWjBtMQswCQYDVQQGEwJTRTESMBAGA1UECgwJWX'
                 'ViaWNvIEFCMSIwIAYDVQQLDBlBdXRoZW50aWNhdG9yIEF0dGVzdGF0aW9uMSY'
                 'wJAYDVQQDDB1ZdWJpY28gVTJGIEVFIFNlcmlhbCA2MTczMDgzNDBZMBMGByqG'
                 'SM49AgEGCCqGSM49AwEHA0IABBmeh5wWLbfcOe5KQqBGFqWzCf7KCS92vglI-'
                 'W1ulcrkzGXNVKBZz73HybMbKx1sGER5wsBh9BiqlUtZaiwc-hejbDBqMCIGCS'
                 'sGAQQBgsQKAgQVMS4zLjYuMS40LjEuNDE0ODIuMS43MBMGCysGAQQBguUcAgE'
                 'BBAQDAgQwMCEGCysGAQQBguUcAQEEBBIEEPormdyeOUJXj5JKMNI8QRgwDAYD'
                 'VR0TAQH_BAIwADANBgkqhkiG9w0BAQsFAAOCAQEAKOuzZ_7R2PDiievKn_bYB'
                 '1fGDprlfLFyjJscOMq7vYTZI32oMawhlJ8PLfwMMWv9sXWzbmOiK7tYDq3KUo'
                 'DQeYQOWh4lcmJaO_uHYDPb-yKpack4uJzhcTWUAKElLZcCqRKT1UUZ6WDdIs6'
                 'KJ-sF6355t1DAAv7ZAWtxHsmtdFAb2RTLvo7ZVxKBt09E6wd85h7LBquFqXJV'
                 'Jn7o45gr9D8Msho4LSNeueTObbKYxAVCUEAjKyth4QzXDGIVvAO36UBxtw4S0'
                 'cR_lmVaLvmdTOVafxtLH_kU7hNtnmEgRxSIZGmIgEQxFmU4ibhkhtnJyf-8k4'
                 'VFNWmzRXRLjKC0NzBFAiEAgx6TM_lyZE6SQx05Wlot-hMK-Cp2p8itseOPNp_'
                 'xtu4CIDpSYJYOb10vM32rjrBNDd-AoKJWbIQRtAwjVEZjXwz6')

    client_data2 = ('eyJjaGFsbGVuZ2UiOiJUdU84MDVPTEQzdXpUNmNnalZ6blRJSjRrZXg4N'
                    'WlremlUTG1aNC1jQnBvIiwib3JpZ2luIjoiaHR0cHM6Ly9sb2NhbGhvc3'
                    'Q6NTAwMCIsInR5cCI6Im5hdmlnYXRvci5pZC5maW5pc2hFbnJvbGxtZW5'
                    '0In0')

    def setUp(self):
        self.setUp_user_realms()
        self.user = User(login='cornelius', resolver=self.resolvername1,
                         realm=self.realm1)
        set_privacyidea_config("u2f.appId", self.app_id)
        # init step 1
        self.token1 = init_token({'type': 'u2f'})
        self.serial1 = self.token1.token.serial
        # finish init step 1
        res = self.token1.get_init_detail()
        # init step 2
        self.token1 = init_token({"type": "u2f",
                                  "serial": self.serial1,
                                  "regdata": self.reg_data1,
                                  "clientdata": self.client_data1},
                                 user=self.user)
        # Token 2
        # init step 1
        self.token2 = init_token({'type': 'u2f'})
        self.serial2 = self.token2.token.serial
        # finish init step 1
        res = self.token2.get_init_detail()
        # init step 2
        self.token2 = init_token({"type": "u2f",
                                  "serial": self.serial2,
                                  "regdata": self.reg_data2,
                                  "clientdata": self.client_data2},
                                 user=self.user)

    def tearDown(self):
        remove_token(serial=self.serial1)
        remove_token(serial=self.serial2)

    # TODO: also test challenge-response with different tokens (u2f + totp)
    def test_01_multiple_token(self):
        set_policy("otppin", scope=SCOPE.AUTH, action="{0!s}=none".format(ACTION.OTPPIN))
        res, reply = check_user_pass(self.user, '')
        self.assertFalse(res)
        self.assertIn('transaction_id', reply, reply)
        tid = reply['transaction_id']
        self.assertIn('multi_challenge', reply, reply)
        self.assertEqual(len(reply['multi_challenge']), 2, reply['multi_challenge'])
        self.assertIn('messages', reply, reply)
        self.assertEqual(len(reply['messages']), 2, reply['messages'])
        # check that the serials of the challenges are different
        chal1 = reply['multi_challenge'][0]
        chal2 = reply['multi_challenge'][1]
        self.assertNotEqual(chal1['serial'], chal2['serial'],
                            reply['multi_challenge'])
        self.assertEqual(chal1['transaction_id'], chal2['transaction_id'],
                         reply['multi_challenge'])
        # Now make sure that the requests contain the same challenge
        self.assertEqual(chal1['attributes']['u2fSignRequest']['challenge'],
                         chal2['attributes']['u2fSignRequest']['challenge'],
                         reply['multi_challenge'])
        # check that we have two challenges in the db with the same challenge
        chals = get_challenges(transaction_id=tid)
        self.assertEqual(len(chals), 2, chals)
        self.assertEqual(chals[0].challenge, chals[1].challenge, chals)

        delete_policy('otppin')



"""Add the DB table password reset in version 2.10

Revision ID: 4023571658f8
Revises: 2ac117d0a6f5
Create Date: 2016-01-07 16:45:35.708951

"""

# revision identifiers, used by Alembic.
revision = '4023571658f8'
down_revision = '2ac117d0a6f5'

from alembic import op
import sqlalchemy as sa
from sqlalchemy.exc import OperationalError, ProgrammingError, InternalError


def upgrade():
    try:
        op.create_table('passwordreset',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('recoverycode', sa.Unicode(length=255), nullable=False),
        sa.Column('username', sa.Unicode(length=64), nullable=False),
        sa.Column('realm', sa.Unicode(length=64), nullable=False),
        sa.Column('resolver', sa.Unicode(length=64), nullable=True),
        sa.Column('email', sa.Unicode(length=255), nullable=True),
        sa.Column('timestamp', sa.DateTime(), nullable=True),
        sa.Column('expiration', sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint('id')
        )
        op.create_index(op.f('ix_passwordreset_realm'), 'passwordreset',
                        ['realm'], unique=False)
        op.create_index(op.f('ix_passwordreset_username'), 'passwordreset',
                        ['username'], unique=False)
    except (OperationalError, ProgrammingError, InternalError) as exx:
        if "duplicate column name" in str(exx.orig).lower():
            print("Good. Table passwordreset already exists.")
        else:
            print(exx)
    except Exception as exx:
        print ("Could not add table 'passwordreset'")
        print (exx)


def downgrade():
    op.drop_index(op.f('ix_passwordreset_username'), table_name='passwordreset')
    op.drop_index(op.f('ix_passwordreset_realm'), table_name='passwordreset')
    op.drop_table('passwordreset')



"""v3.8: Sequence for cusstomuserattribute, tokengroup and tokentokengroup

Revision ID: a28f2733897b
Revises: 89e57ed16379
Create Date: 2022-11-24 11:05:42.572284

"""

# revision identifiers, used by Alembic.
revision = 'a28f2733897b'
down_revision = '89e57ed16379'

from alembic import op, context
import sqlalchemy as sa
from sqlalchemy.schema import Sequence, CreateSequence

def dialect_supports_sequences():
    migration_context = context.get_context()
    return migration_context.dialect.supports_sequences


def create_seq(seq):
    if dialect_supports_sequences():
        op.execute(CreateSequence(seq))


def upgrade():
    try:
        seq = Sequence('customuserattribute_seq')
        create_seq(seq)
    except Exception as exx:
        print(exx)

    try:
        seq = Sequence('tokengroup_seq')
        create_seq(seq)
    except Exception as exx:
        print(exx)

    try:
        seq = Sequence('tokentokengroup_seq')
        create_seq(seq)
    except Exception as exx:
        print(exx)


def downgrade():
    pass



"""v3.11: Fix content type in EventHandlerOption table

Revision ID: 7301d5130c3a
Revises: 69e7817b9863
Create Date: 2024-11-29 12:03:46.090125

"""
from alembic import op
from sqlalchemy import orm, Unicode
from sqlalchemy.sql import table, column

# revision identifiers, used by Alembic.
revision = '7301d5130c3a'
down_revision = 'eac770c0bbed'

eventhandleroption = table("eventhandleroption", column("Key", Unicode(255)), column("Value", Unicode(2000)))


def upgrade():
    bind = op.get_bind()
    session = orm.Session(bind=bind)
    try:
        op.execute(eventhandleroption.update()
                   .where(eventhandleroption.c.Key == op.inline_literal("content_type"))
                   .where(eventhandleroption.c.Value == op.inline_literal("urlendcode"))
                   .values({"Value": op.inline_literal("urlencoded")}))
    except Exception as e:
        print("Failed to update urlendcode to urlencoded in eventhandleroption table.")
        print(e)
        session.rollback()
        raise

    finally:
        session.commit()


def downgrade():
    bind = op.get_bind()
    session = orm.Session(bind=bind)
    try:
        op.execute(eventhandleroption.update()
                   .where(eventhandleroption.c.Key == op.inline_literal("content_type"))
                   .where(eventhandleroption.c.Value == op.inline_literal("urlencoded"))
                   .values({"Value": op.inline_literal("urlendcode")}))
    except Exception as e:
        print("Failed to revert urlencoded to urlendcode in eventhandleroption table.")
        print(e)
        session.rollback()
        raise

    finally:
        session.commit()



from .pkcs11mock import PKCS11Mock
from privacyidea.lib.resolver import save_resolver
from privacyidea.lib.security.aeshsm import AESHardwareSecurityModule
from tests import ldap3mock
from tests.base import MyApiTestCase
from tests.test_api_validate import LDAPDirectory


class APIHealthcheckTestCase(MyApiTestCase):
    def test_livez(self):
        with self.app.test_request_context('/healthz/livez', method='GET'):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, f"Expected status code 200, got {res.status_code}")
            result = res.json.get("result")
            self.assertIsNotNone(result, "Expected 'result' in response JSON, but got None")
            self.assertIn("value", result, f"Expected 'value' key in result, got {result}")
            self.assertEqual(result.get("value").get("status"), "OK",
                             f"Expected 'OK' as value, got {result.get('value')}")

    def test_startupz(self):
        def check_status(expected_status_code, expected_ready_status, app_ready_value):
            self.app.config['APP_READY'] = app_ready_value
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, expected_status_code,
                             f"Expected status code {expected_status_code}, got {res.status_code}")
            result = res.json.get("result")
            self.assertIsNotNone(result, "Expected JSON result, got None")
            self.assertIn("value", result, f"Expected 'value' in result, got {result}")
            self.assertEqual(result.get("value").get("status"), expected_ready_status,
                             f"Expected status '{expected_ready_status}', got {result.get('value').get('status')}")

        with self.app.test_request_context('/healthz/startupz', method='GET'):
            check_status(503, "not started", False)
            check_status(200, "started", True)

    def test_readyz_and_healthz(self):
        def check_status(expected_status_code, expected_ready_status, app_ready_value):
            with PKCS11Mock():
                hsm = AESHardwareSecurityModule({
                    "module": "testmodule",
                    "password": "test123!"
                })
                self.assertIsNotNone(hsm)
                self.app.config['APP_READY'] = app_ready_value
                res = self.app.full_dispatch_request()
                self.assertEqual(res.status_code, expected_status_code,
                                 f"Expected status code {expected_status_code}, got {res.status_code}")
                result = res.json.get("result")
                self.assertIsNotNone(result, "Expected JSON result, got None")
                self.assertIn("value", result, f"Expected 'value' in result, got {result}")
                self.assertEqual(result.get("value").get("status"), expected_ready_status,
                                 f"Expected status '{expected_ready_status}', got {result.get('value').get('status')}")

        with self.app.test_request_context('/healthz/readyz', method='GET'):
            check_status(503, "not ready", False)
            check_status(200, "ready", True)
        with self.app.test_request_context('/healthz/', method='GET'):
            check_status(503, "not ready", False)
            check_status(200, "ready", True)

    @ldap3mock.activate
    def test_resolversz(self):
        def check_resolvers(res, expected_status_code, expected_status, ldap_expected_status,
                            sql_expected_status):
            assert res.status_code == expected_status_code, (f"Expected status code {expected_status_code}, "
                                                             f"got {res.status_code}")
            result = res.json.get("result")
            assert result is not None, "Expected JSON result, got None"
            assert result.get("value").get(
                "status") == expected_status, (f"Expected '{expected_status}' status, "
                                               f"got {result.get('value').get('status')}")

            ldap_resolvers = result.get("value").get("ldapresolver")
            sql_resolvers = result.get("value").get("sqlresolver")

            assert ldap_resolvers is not None, "Expected 'ldapresolver' in result, but got None"
            assert sql_resolvers is not None, "Expected 'sqlresolver' in result, but got None"

            assert all(status == ldap_expected_status for status in ldap_resolvers.values()), \
                f"At least one LDAP resolver does not have '{ldap_expected_status}' status."
            assert all(status == sql_expected_status for status in sql_resolvers.values()), \
                f"At least one SQL resolver does not have '{sql_expected_status}' status."

        ldap3mock.setLDAPDirectory(LDAPDirectory)

        with self.app.test_request_context('/healthz/resolversz', method='GET'):
            res = self.app.full_dispatch_request()
            check_resolvers(res, 200, "OK",
                            ldap_expected_status="fail",
                            sql_expected_status="fail")

            ldapr = save_resolver({
                'LDAPURI': 'ldap://localhost',
                'LDAPBASE': 'o=test',
                'BINDDN': 'cn=manager,ou=example,o=test',
                'BINDPW': 'ldaptest',
                'LOGINNAMEATTRIBUTE': 'cn',
                'LDAPSEARCHFILTER': '(cn=*)',
                'MULTIVALUEATTRIBUTES': '["groups"]',
                'USERINFO': '{ "username": "cn",'
                            '"phone" : "telephoneNumber", '
                            '"mobile" : "mobile", '
                            '"email" : "mail", '
                            '"surname" : "sn", '
                            '"groups": "memberOf", '
                            '"givenname" : "givenName" }',
                'UIDTYPE': 'DN',
                "resolver": "test_ldapresolver",
                "type": "ldapresolver"
            })

            sqlr = save_resolver({
                'Driver': 'sqlite',
                'Server': '/tests/testdata/',
                'Database': "testuser.sqlite",
                'Table': 'users',
                'Encoding': 'utf8',
                'Editable': True,
                'Map': '{ "username": "username", \
                             "userid" : "id", \
                             "email" : "email", \
                             "surname" : "name", \
                             "givenname" : "givenname", \
                             "password" : "password", \
                             "phone": "phone", \
                             "mobile": "mobile"}',
                'resolver': "test_sqlresolver",
                'type': "sqlresolver"
            })

            self.assertGreater(ldapr, 0, "LDAP resolver creation failed.")
            self.assertGreater(sqlr, 0, "SQL resolver creation failed.")

            res = self.app.full_dispatch_request()
            check_resolvers(res, 200, "OK",
                            ldap_expected_status="OK",
                            sql_expected_status="OK")



import logging
import re

log = logging.getLogger(__name__)


def validate_email(email):
    """
    Verify if the email is a gmail address.
    :param email: email address
    :return: True if valid, False otherwise
    """
    # regular expression for email validation
    regex = r'^[-\w\.]+@gmail.com$'
    if re.search(regex, email):
        return True
    else:
        return False



#  2021-10-05 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Allow to set a default slot number
#  2017-09-25 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Add reinitialization of the PKCS11 module
#  2016-09-01 Mathias Brossard <mathias@axiadids.com>
#             Alternate PKCS11 Security Module
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#

import logging
import datetime
from privacyidea.lib.security.default import (SecurityModule,
                                              int_list_to_bytestring)
from privacyidea.lib.error import HSMException
from privacyidea.lib.crypto import get_alphanum_str

__doc__ = """
This is a PKCS11 Security module that encrypts and decrypts the data on a
HSM that is connected via PKCS11. This alternate version relies on AES keys.
"""

log = logging.getLogger(__name__)

MAX_RETRIES = 5

try:
    import PyKCS11
except ImportError:
    log.info("The python module PyKCS11 is not available. "
             "So we can not use the PKCS11 security module.")


class AESHardwareSecurityModule(SecurityModule):  # pragma: no cover

    def __init__(self, config=None):
        """
        Initialize the PKCS11 Security Module.
        The configuration needs to contain the pkcs11 module and the ID of the key.

        {"module": "/usr/lib/hsm_pkcs11.so", "slot": 42, "key_label": "privacyidea"}

        The HSM is not directly ready, since the HSM is protected by a password.
        The function setup_module({"password": "HSM User password"}) needs to be called.

        :param config: contains the HSM configuration
        :type config: dict

        :return: The Security Module object
        """
        config = config or {}
        self.name = "HSM"
        self.config = config
        # Initially, we might be missing a password
        self.is_ready = False

        if "module" not in config:
            log.error("No PKCS11 module defined!")
            raise HSMException("No PKCS11 module defined.")

        label_prefix = config.get("key_label", "privacyidea")
        self.key_labels = {}
        for k in self.mapping:
            l = config.get(("key_label_{0!s}".format(k)))
            l = ('{0!s}_{1!s}'.format(label_prefix, k)) if l is None else l
            self.key_labels[k] = l

        log.debug("Setting key labels: {0!s}".format(self.key_labels))
        # convert the slot to int
        self.slot = int(config.get("slot", 1))
        log.debug("Setting slot: {0!s}".format(self.slot))
        self.password = config.get("password")
        log.debug("Setting a password: {0!s}".format(bool(self.password)))
        self.module = config.get("module")
        log.debug("Setting the modules: {0!s}".format(self.module))
        self.max_retries = config.get("max_retries", MAX_RETRIES)
        log.debug("Setting max retries: {0!s}".format(self.max_retries))
        self.session = None
        self.session_start_time = datetime.datetime.now()
        self.session_lastused_time = datetime.datetime.now()
        self.key_handles = {}

        self.initialize_hsm()

    def initialize_hsm(self):
        """
        Initialize the HSM:
        * initialize PKCS11 library
        * login to HSM
        * get session
        :return:
        """
        self.pkcs11 = PyKCS11.PyKCS11Lib()
        self.pkcs11.load(self.module)
        self.pkcs11.lib.C_Initialize()
        if self.password:
            self._login()

    def setup_module(self, params):
        """
        callback, which is called during the runtime to initialze the
        security module.

        Here the password for the PKCS11 HSM can be provided

           {"password": "top secreT"}

        :param params: The password for the HSM
        :type  params: dict

        :return: -
        """
        if "password" in params:
            self.password = str(params.get("password"))
        else:
            raise HSMException("missing password")
        self._login()
        return self.is_ready

    def _login(self):
        slotlist = self.pkcs11.getSlotList()
        log.debug("Found the slots: {0!s}".format(slotlist))
        if not len(slotlist):
            raise HSMException("No HSM connected. No slots found.")
        if self.slot == -1 and len(slotlist) == 1:
            # Use the first and only slot
            self.slot = slotlist[0]
        if self.slot not in slotlist:
            raise HSMException("Slot {0:d} not present".format(self.slot))

        slotinfo = self.pkcs11.getSlotInfo(self.slot)
        log.debug("Setting up '{}'".format(slotinfo.slotDescription))

        # Before starting the session, we log the old session time usage
        log.debug("Starting new session now. The old session started {0!s} seconds ago.".format(
            datetime.datetime.now() - self.session_start_time))
        log.debug("Starting new session now. The old session was used {0!s} seconds ago.".format(
            datetime.datetime.now() - self.session_lastused_time))
        # If the HSM is not connected at this point, it will fail
        self.session = self.pkcs11.openSession(slot=self.slot)
        self.session_start_time = datetime.datetime.now()

        log.debug("Logging on to '{}'".format(slotinfo.slotDescription))
        self.session.login(self.password)

        for k in self.mapping:
            label = self.key_labels[k]
            objs = self.session.findObjects([(PyKCS11.CKA_CLASS, PyKCS11.CKO_SECRET_KEY),
                                             (PyKCS11.CKA_LABEL, label)])
            log.debug("Loading '{}' key with label '{}'".format(k, label))
            if objs:
                self.key_handles[self.mapping[k]] = objs[0]

        # self.session.logout()
        log.debug("Successfully setup the security module.")
        self.is_ready = True

    def random(self, length):
        """
        Return a random bytestring
        :param length: length of the random bytestring
        :rtype bytes
        """
        retries = 0
        while True:
            try:
                r_integers = self.session.generateRandom(length)
                self.session_lastused_time = datetime.datetime.now()
                break
            except PyKCS11.PyKCS11Error as exx:
                log.warning("Generate Random failed: {0!s}".format(exx))
                # If we get an CKR_SESSION_HANDLE_INVALID error code, we free
                # memory, session and handles and retry
                if exx.value == PyKCS11.CKR_SESSION_HANDLE_INVALID:
                    self.pkcs11.lib.C_Finalize()
                    self.initialize_hsm()
                    retries += 1
                    if retries > self.max_retries:
                        raise HSMException("Failed to generate random number after multiple retries.")
                else:
                    raise HSMException("HSM random number generation failed "
                                       "with {0!s}".format(exx))

        # convert the array of the random integers to a string
        return int_list_to_bytestring(r_integers)

    def encrypt(self, data, iv, key_id=SecurityModule.TOKEN_KEY):
        """

        :rtype: bytes
        """
        log.debug("Encrypting {} bytes with key {}".format(len(data), key_id))
        m = PyKCS11.Mechanism(PyKCS11.CKM_AES_CBC_PAD, iv)
        retries = 0
        while True:
            try:
                k = self.key_handles[key_id]
                r = self.session.encrypt(k, bytes(data), m)
                self.session_lastused_time = datetime.datetime.now()
                break
            except PyKCS11.PyKCS11Error as exx:
                log.warning("Encryption failed: {0!s}".format(exx))
                # If we get an CKR_SESSION_HANDLE_INVALID error code, we free
                # memory, session and handles and retry
                if exx.value == PyKCS11.CKR_SESSION_HANDLE_INVALID:
                    self.pkcs11.lib.C_Finalize()
                    self.initialize_hsm()
                    retries += 1
                    if retries > self.max_retries:
                        raise HSMException("Failed to encrypt after multiple retries")
                else:
                    raise HSMException("HSM encryption failed with {0!s}".format(exx))

        return int_list_to_bytestring(r)

    def decrypt(self, enc_data, iv, key_id=SecurityModule.TOKEN_KEY):
        """

        :rtype bytes
        """
        # we keep this for legacy reasons, even though it hasn't worked anyway
        if len(enc_data) == 0:
            return bytes()
        log.debug("Decrypting {} bytes with key {}".format(len(enc_data), key_id))
        m = PyKCS11.Mechanism(PyKCS11.CKM_AES_CBC_PAD, iv)
        start = datetime.datetime.now()
        retries = 0
        while True:
            try:
                k = self.key_handles[key_id]
                r = self.session.decrypt(k, bytes(enc_data), m)
                self.session_lastused_time = datetime.datetime.now()
                break
            except PyKCS11.PyKCS11Error as exx:
                log.warning("Decryption failed: {0!s}".format(exx))
                # If we get an CKR_SESSION_HANDLE_INVALID error code, we free
                # memory, session and handles and retry
                if exx.value == PyKCS11.CKR_SESSION_HANDLE_INVALID:
                    self.pkcs11.lib.C_Finalize()
                    self.initialize_hsm()
                    retries += 1
                    if retries > self.max_retries:
                        td = datetime.datetime.now() - start
                        log.warning("Decryption finally failed: {0!s}. Time taken: {1!s}.".format(exx, td))
                        raise HSMException("Failed to decrypt after multiple retries.")
                else:
                    raise HSMException("HSM decrypt failed with {0!s}".format(exx))

        if retries > 0:
            td = datetime.datetime.now() - start
            log.warning("Decryption after {0!s} retries successful. Time taken: {1!s}.".format(retries, td))
        return int_list_to_bytestring(r)

    def create_keys(self):
        """
        Connect to the HSM and create the encryption keys.
        The HSM connection should already be configured in pi.cfg.

        We will create new keys with new key labels
        :return: a dictionary of the created key labels
        """
        # We need a new read/write session
        session = self.pkcs11.openSession(self.slot,
                                          PyKCS11.CKF_SERIAL_SESSION | PyKCS11.CKF_RW_SESSION)
        # We need to logout, otherwise we get CKR_USER_ALREADY_LOGGED_IN
        session.logout()
        session.login(self.password)

        key_labels = {"token": "",
                      "config": "",
                      "value": ""}

        for kl in key_labels.keys():
            label = "{0!s}_{1!s}".format(kl, get_alphanum_str())
            aesTemplate = [
                (PyKCS11.CKA_CLASS, PyKCS11.CKO_SECRET_KEY),
                (PyKCS11.CKA_KEY_TYPE, PyKCS11.CKK_AES),
                (PyKCS11.CKA_VALUE_LEN, 32),
                (PyKCS11.CKA_LABEL, label),
                (PyKCS11.CKA_TOKEN, PyKCS11.CK_TRUE),
                (PyKCS11.CKA_PRIVATE, True),
                (PyKCS11.CKA_SENSITIVE, True),
                (PyKCS11.CKA_ENCRYPT, True),
                (PyKCS11.CKA_DECRYPT, True),
                (PyKCS11.CKA_TOKEN, True),
                (PyKCS11.CKA_WRAP, True),
                (PyKCS11.CKA_UNWRAP, True),
                (PyKCS11.CKA_EXTRACTABLE, False)
            ]
            aesKey = session.generateKey(aesTemplate)
            key_labels[kl] = label

        session.logout()
        session.closeSession()

        return key_labels


if __name__ == "__main__":  # pragma: no cover
    logging.basicConfig()
    log.setLevel(logging.INFO)
    # log.setLevel(logging.DEBUG)

    module = "/usr/local/opt/pkcs11/lib/pkcs11-token.so"

    p = AESHardwareSecurityModule({"module": module, "slot": 2,
                                   "key_label": "privacyidea",
                                   "password": "12345678"})

    # password
    password = "topSekr3t" * 16
    crypted = p.encrypt_password(password)
    text = p.decrypt_password(crypted)
    assert(text == password)  # nosec B101 # This is actually a test
    log.info("password encrypt/decrypt test successful")

    # pin
    password = "topSekr3t"  # nosec B105 # used for testing
    crypted = p.encrypt_pin(password)
    text = p.decrypt_pin(crypted)
    assert (text == password)  # nosec B101 # This is actually a test
    log.info("pin encrypt/decrypt test successful")

    p = AESHardwareSecurityModule({"module": module, "slot": 2,
                                   "key_label": "privacyidea"})
    p.setup_module({"password": "12345678"})

    # random
    tmp_iv = p.random(16)
    plain = p.random(128)
    log.info("random test successful")

    # generic encrypt / decrypt
    cipher = p.encrypt(plain, tmp_iv)
    assert (plain != cipher)  # nosec B101 # This is actually a test
    text = p.decrypt(cipher, tmp_iv)
    assert (text == plain)  # nosec B101 # This is actually a test
    log.info("generic encrypt/decrypt test successful")



"""add dedicated adminuser to policies

Revision ID: a7e91b18a460
Revises: 0c7123345224
Create Date: 2020-01-29 13:42:15.390923

"""

# revision identifiers, used by Alembic.
revision = 'a7e91b18a460'
down_revision = '0c7123345224'

from alembic import op
import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()


# Use table definition from the point where the changes took place. If we use
# the table definition from models.py we might run into problems
class Policy(Base):
    __tablename__ = "policy"
    __table_args__ = {'mysql_row_format': 'DYNAMIC'}
    id = sa.Column(sa.Integer, sa.Sequence("policy_seq"), primary_key=True)
    active = sa.Column(sa.Boolean, default=True)
    check_all_resolvers = sa.Column(sa.Boolean, default=False)
    name = sa.Column(sa.Unicode(64), unique=True, nullable=False)
    scope = sa.Column(sa.Unicode(32), nullable=False)
    action = sa.Column(sa.Unicode(2000), default=u"")
    realm = sa.Column(sa.Unicode(256), default=u"")
    adminrealm = sa.Column(sa.Unicode(256), default=u"")
    adminuser = sa.Column(sa.Unicode(256), default=u"")
    resolver = sa.Column(sa.Unicode(256), default=u"")
    user = sa.Column(sa.Unicode(256), default=u"")
    client = sa.Column(sa.Unicode(256), default=u"")
    time = sa.Column(sa.Unicode(64), default=u"")
    priority = sa.Column(sa.Integer, default=1, nullable=False)


def upgrade():
    try:
        op.add_column('policy', sa.Column('adminuser', sa.Unicode(length=256), nullable=True))
    except Exception as exx:
        print('Adding of column "adminuser" in table policy failed: {!r}'.format(exx))
        print('This is expected behavior if this column already exists.')

    # Now that we added the column in the table, we can move the "user" from admin-policies to
    # the "adminuser" column
    bind = op.get_bind()
    session = orm.Session(bind=bind)
    pol_name = None
    try:
        for policy in session.query(Policy).filter(Policy.user != u"", Policy.scope == u"admin"):
            pol_name = policy.name
            # move the "user" to the "adminuser"
            policy.adminuser = policy.user
            policy.user = u""
        session.commit()
    except Exception as exx:
        session.rollback()
        print("Failed to migrate column adminuser in policies due to error in "
              "policy '{0!s}'.".format(pol_name))
        print(exx)


def downgrade():
    op.drop_column('policy', 'adminuser')



#  2015-06-30 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Initial writup
#
# License:  AGPLv3
# (c) 2015. Cornelius Kölbel
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
__doc__ = """This module provides the PIN Handling base class.
In case of enrolling a token, a PIN Handling class can be used to
send the PIN via Email, call an external program or print a letter.

This module is not tested explicitly.
It is tested in conjunction with the policy decorator init_random_pin in
tests/test_api_lib_policy.py
"""
import logging
log = logging.getLogger(__name__)


class PinHandler(object):
    """
    A PinHandler Class is responsible for handling the OTP PIN during
    enrollment.

    It receives the necessary data like
      * the PIN
      * the serial number of the token
      * the username
      * all other user data:

        *  given name, surname
        *  email address
        *  telephone
        *  mobile (if the module would deliver via SMS)
      * the administrator name (who enrolled the token)
    """
    def __init__(self, options=None):
        pass

    def send(self, pin, serial, user, tokentype=None, logged_in_user=None,
             userdata=None, options=None):
        """

        :param pin: The PIN in cleartext
        :param user: the owner of the token
        :type user: user object
        :param tokentype: the type of the token
        :type tokentype: basestring
        :param logged_in_user: The logged in user, who enrolled the token
        :type logged_in_user: dict
        :param userdata: Handler-specific user data like email, mobile...
        :type userdata: dict
        :param options: Handler-specific additional options
        :type options: dict
        :return: True in case of success
        :rtype: bool
        """
        # The most simple way of handling a random PIN! ;-)
        log.info("handling pin {0!r} for token {1!s} of user {2!r}".format(pin, serial,
                                                              user))
        log.info("The token was enrolled by {0!r}@{1!s}".format(logged_in_user.get("username"), logged_in_user.get("realm")))
        return True



"""v3.9: Add table serviceid

Revision ID: 4a0aec37e7cf
Revises: 006d4747f858
Create Date: 2023-03-15 16:14:26.204225

"""

# revision identifiers, used by Alembic.
revision = '4a0aec37e7cf'
down_revision = '006d4747f858'

from alembic import op, context
import sqlalchemy as sa
from sqlalchemy.schema import Sequence, CreateSequence


def dialect_supports_sequences():
    migration_context = context.get_context()
    return migration_context.dialect.supports_sequences


def create_seq(seq):
    if dialect_supports_sequences():
        op.execute(CreateSequence(seq))


def upgrade():
    try:
        seq = Sequence('serviceid_seq')
        try:
            create_seq(seq)
        except Exception as _e:
            pass
        op.create_table('serviceid',
        sa.Column('id', sa.Integer(), seq, nullable=False),
        sa.Column('name', sa.Unicode(length=255), nullable=False),
        sa.Column('Description', sa.Unicode(length=2000), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('name'),
        mysql_row_format='DYNAMIC'
        )
    except Exception as exx:
        print("Could not add table 'serviceid' - probably already exists!")
        print(exx)


def downgrade():
    op.drop_table('serviceid')



"""Add table eventcounter

Revision ID: 145ce80decd
Revises: 49a04e560d96
Create Date: 2018-03-01 12:22:04.523615

"""

# revision identifiers, used by Alembic.
revision = '145ce80decd'
down_revision = '49a04e560d96'

from alembic import op
import sqlalchemy as sa


def upgrade():
    try:
        op.create_table('eventcounter',
                        sa.Column('counter_name', sa.Unicode(length=80), nullable=False),
                        sa.Column('counter_value', sa.Integer(), nullable=True),
                        sa.PrimaryKeyConstraint('counter_name')
                        )
    except Exception as exx:
        print("Could not create table eventcounter. Probably already exists!")
        print (exx)


def downgrade():
    op.drop_table('eventcounter')



"""Store privacyIDEA node in eventcounter table

Revision ID: d756b34061ff
Revises: 3d7f8b29cbb1
Create Date: 2019-09-02 13:59:24.244529

"""
# revision identifiers, used by Alembic.
from sqlalchemy import orm
from sqlalchemy.sql.ddl import CreateSequence

from privacyidea.lib.config import get_privacyidea_node

revision = 'd756b34061ff'
down_revision = '3d7f8b29cbb1'

from alembic import op, context
import sqlalchemy as sa
from sqlalchemy.ext.declarative import declarative_base


Base = declarative_base()


class OldEventCounter(Base):
    __tablename__ = 'eventcounter'
    counter_name = sa.Column(sa.Unicode(80), nullable=False, primary_key=True)
    counter_value = sa.Column(sa.Integer, default=0)
    __table_args__ = {'mysql_row_format': 'DYNAMIC'}


class NewEventCounter(Base):
    __tablename__ = 'eventcounter_new'
    id = sa.Column(sa.Integer, sa.Sequence("eventcounter_seq"), primary_key=True)
    counter_name = sa.Column(sa.Unicode(80), nullable=False)
    counter_value = sa.Column(sa.Integer, default=0)
    node = sa.Column(sa.Unicode(255), nullable=False)
    __table_args__ = (sa.UniqueConstraint('counter_name',
                                          'node',
                                          name='evctr_1'),
                      {'mysql_row_format': 'DYNAMIC'})


def dialect_supports_sequences():
    migration_context = context.get_context()
    return migration_context.dialect.supports_sequences


def create_seq(seq):
    if dialect_supports_sequences():
        op.execute(CreateSequence(seq))


def upgrade():
    bind = op.get_bind()
    session = orm.Session(bind=bind)
    try:
        # Step 1: Create sequence on Postgres
        seq = sa.Sequence('eventcounter_seq')
        try:
            create_seq(seq)
        except Exception as _e:
            pass
        # Step 2: Create new eventcounter_new table
        op.create_table('eventcounter_new',
                        sa.Column("id", sa.Integer, sa.Sequence("eventcounter_seq"), primary_key=True),
                        sa.Column("counter_name", sa.Unicode(80), nullable=False),
                        sa.Column("counter_value", sa.Integer, default=0),
                        sa.Column("node", sa.Unicode(255), nullable=False),
                        sa.UniqueConstraint('counter_name', 'node', name='evctr_1'),
                        mysql_row_format='DYNAMIC'
                        )
        # Step 3: Migrate data from eventcounter to eventcounter_new
        node = get_privacyidea_node()
        for old_ctr in session.query(OldEventCounter).all():
            new_ctr = NewEventCounter(counter_name=old_ctr.counter_name,
                                      counter_value=old_ctr.counter_value,
                                      node=node)
            session.add(new_ctr)
            print("Migrating counter {!r}={} on node={!r} ...".format(new_ctr.counter_name,
                                                                      new_ctr.counter_value,
                                                                      node))
        session.commit()
        # Step 4: Remove eventcounter
        op.drop_table("eventcounter")
        op.rename_table("eventcounter_new", "eventcounter")
    except Exception as exx:
        session.rollback()
        print("Could not migrate table 'eventcounter'")
        print (exx)

def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint('evctr_1', 'eventcounter', type_='unique')
    op.drop_column('eventcounter', 'node')
    op.drop_column('eventcounter', 'id')
    # ### end Alembic commands ###



from .base import MyApiTestCase
from privacyidea.lib.user import (User)
from privacyidea.lib.token import init_token
from privacyidea.lib.policy import SCOPE, ACTION, set_policy
from privacyidea.lib.machine import attach_token


class OfflinePassNoTokenTestCase(MyApiTestCase):
    """
    Ensure that a user with no tokens do not receive any offline values
    https://github.com/privacyidea/privacyidea/issues/3333
    """

    def test_00_setup(self):
        self.setUp_user_realms()

    def test_01_create_tokens(self):
        # create token for 1st user

        userA = User("shadow", self.realm1, self.resolvername1)
        userB = User("cornelius", self.realm1, self.resolvername1)
        userC = User("usernotoken", self.realm1, self.resolvername1)
        tokA = init_token({"otpkey": self.otpkey, "serial": "tokA"},
                          user=userA)
        tokB = init_token({"otpkey": self.otpkey, "serial": "tokB"},
                          user=userB)
        # attach tokens to offline
        attach_token("tokA", "offline")
        attach_token("tokB", "offline")
        # set passonnotoken
        set_policy(name="pol1", scope=SCOPE.AUTH, action=ACTION.PASSNOTOKEN)

        # Validate userB
        with self.app.test_request_context('/validate/check',
                                           method='POST',
                                           data={"user": "cornelius",
                                                 "pass": self.valid_otp_values[1]},
                                           environ_base={'REMOTE_ADDR': '192.168.0.2'}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            data = res.json
            result = data.get("result")
            self.assertEqual("ACCEPT", result.get("authentication"))
            detail = data.get("detail")
            self.assertEqual("matching 1 tokens", detail.get("message"))
            auth_items = data.get("auth_items")
            self.assertIn("offline", auth_items)
            offline = auth_items.get("offline")
            # One offline entry
            self.assertEqual(1, len(offline))
            self.assertEqual("cornelius", offline[0].get("username"))
            serial = offline[0].get("serial")
            self.assertEqual("tokB", serial)
            refilltoken = offline[0].get("refilltoken")

        # Test refill
        with self.app.test_request_context('/validate/offlinerefill',
                                           method='POST',
                                           data={"serial": serial,
                                                 "refilltoken": refilltoken,
                                                 "pass": self.valid_otp_values[3]},
                                           environ_base={'REMOTE_ADDR': '192.168.0.2'}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            data = res.json
            auth_items = data.get("auth_items")
            self.assertIn("offline", auth_items)
            offline = auth_items.get("offline")
            self.assertEqual(1, len(offline))
            self.assertIn("serial", offline[0])

        # PassOnNoToken userC
        with self.app.test_request_context('/validate/check',
                                           method='POST',
                                           data={"user": "usernotoken",
                                                 "pass": "sthelse"},
                                           environ_base={'REMOTE_ADDR': '192.168.0.2'}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            data = res.json
            result = data.get("result")
            self.assertTrue(result.get("value"))
            detail = data.get("detail")
            self.assertEqual("user has no token, accepted due to 'pol1'",
                             detail.get("message"))
            # No AuthItems for this user
            self.assertNotIn("auth_items", data)

        # Validate userA
        with self.app.test_request_context('/validate/check',
                                           method='POST',
                                           data={"user": "shadow",
                                                 "pass": self.valid_otp_values[1]},
                                           environ_base={'REMOTE_ADDR': '192.168.0.2'}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            data = res.json
            result = data.get("result")
            self.assertEqual("ACCEPT", result.get("authentication"))
            detail = data.get("detail")
            self.assertEqual("matching 1 tokens", detail.get("message"))
            auth_items = data.get("auth_items")
            self.assertIn("offline", auth_items)
            offline = auth_items.get("offline")
            # One offline entry
            self.assertEqual(1, len(offline))
            self.assertEqual("shadow", offline[0].get("username"))



#  2018-08-06 Paul Lettich <paul.lettich@netknights.it>
#
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
import logging

from privacyidea.lib.utils import is_true
from privacyidea.lib.tokenclass import TOKENKIND
from privacyidea.lib.token import get_tokens
from privacyidea.lib.monitoringstats import write_stats
from privacyidea.lib.subscriptions import get_users_with_active_tokens
from privacyidea.lib.task.base import BaseTask
# from privacyidea.lib.user import get_user_list
from privacyidea.lib import _

__doc__ = """This is a statistics task which collects simple statistics from the database.
If You want to add more statistic points, simply add them to the options method and add a
corresponding property function (beginning with a '_').
The entry in the monitoringstats table will have the same key as the property name."""

log = logging.getLogger(__name__)


class SimpleStatsTask(BaseTask):
    identifier = "SimpleStats"
    description = "Collect simple statistics"

    @property
    def options(self):
        return {
            "total_tokens": {
                "type": "bool",
                "description": _("Total number of tokens")},
            "hardware_tokens": {
                "type": "bool",
                "description": _("Total number of hardware tokens")},
            "software_tokens": {
                "type": "bool",
                "description": _("Total number of software tokens")},
            "unassigned_hardware_tokens": {
                "type": "bool",
                "description": _("Number of hardware tokens not assigned to a user")},
            "assigned_tokens": {
                "type": "bool",
                "description": _("Number of tokens assigned to users")},
            "user_with_token": {
                "type": "bool",
                "description": _("Number of users with tokens assigned")}
            }

    @property
    def _user_with_token(self):
        return get_users_with_active_tokens()

    @property
    def _total_tokens(self):
        return get_tokens(count=True)

    @property
    def _hardware_tokens(self):
        return get_tokens(count=True, tokeninfo={'tokenkind': TOKENKIND.HARDWARE})

    @property
    def _software_tokens(self):
        return get_tokens(count=True, tokeninfo={'tokenkind': TOKENKIND.SOFTWARE})

    @property
    def _unassigned_hardware_tokens(self):
        return get_tokens(count=True, tokeninfo={'tokenkind': 'hardware'}, assigned=False)

    @property
    def _assigned_tokens(self):
        return get_tokens(count=True, assigned=True)

    def do(self, params):
        for opt in self.options.keys():
            if is_true(params.get(opt)):
                log.debug("Got param {0}".format(opt))
                write_stats(opt, getattr(self, '_' + opt))

        return True



# SPDX-FileCopyrightText: (C) 2024 Jona-Samuel Höhmann <jona-samuel.hoehmann@netknights.it>
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# Info: https://privacyidea.org
#
# 2024-11-11 Jona-Samuel Höhmann <jona-samuel.hoehmann@netknights.it>
#            New pi-token-janitor script
# 2023-11-03 Jona-Samuel Höhmann <jona-samuel.hoehmann@netknights.it>
#            Migrate to click
# 2020-11-11 Timo Sturm <timo.sturm@netknights.it>
#            Select how to validate PSKC imports
# 2018-02-21 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#            Allow to import PSKC file
# 2017-11-21 Cornelius Kölbel <corenlius.koelbel@netknights.it>
#            export to CSV including usernames
# 2017-10-18 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#            Add token export (HOTP and TOTP) to PSKC
# 2017-05-02 Friedrich Weber <friedrich.weber@netknights.it>
#            Improve token matching
# 2017-04-25 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#
# Copyright (c) 2017, Cornelius Kölbel
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE

import click
from flask.cli import with_appcontext
from yaml import safe_load as yaml_safe_load
from privacyidea.lib.token import get_tokens
import sys


@click.command("update")
@click.argument('yaml_file', type=click.File())
@with_appcontext
def updatetokens(yaml_file):
    """
    Update existing tokens in the privacyIDEA system. You must specify a YAML
    file with the tokendata.
    Can be used to reencrypt data, when changing the encryption key.
    """
    click.echo("Loading YAML data. This may take a while.")
    token_list = yaml_safe_load(yaml_file.read())
    for tok in token_list:
        del (tok["owner"])
        serial = tok.get("serial")
        tok_objects = get_tokens(serial=serial)
        if len(tok_objects) == 0:
            sys.stderr.write(f"\nCan not find token {serial}. Not updating.\n")
        else:
            click.echo(f"Updating token {serial}.")
            try:
                tok_objects[0].update(tok)
            except Exception as e:
                click.echo(f"\nFailed to update token {serial} ({e}).", err=True)



#  privacyIDEA is a fork of LinOTP
#
#  2014-12-07 Cornelius Kölbel <cornelius@privacyidea.org>
#
#  Copyright (C) 2014 Cornelius Kölbel
#  License:  AGPLv3
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
This is a helper module for the challenges database table.
It is used by the lib.tokenclass

The method is tested in test_lib_challenges
"""

import logging
from .log import log_with
from ..models import Challenge, db

log = logging.getLogger(__name__)


@log_with(log)
def get_challenges(serial=None, transaction_id=None, challenge=None):
    """
    This returns a list of database challenge objects.

    :param serial: challenges for this very serial number
    :param transaction_id: challenges with this very transaction id
    :param challenge: The challenge to be found
    :return: list of objects
    """
    sql_query = Challenge.query

    if serial is not None:
        sql_query = sql_query.filter(Challenge.serial == serial)

    if transaction_id is not None:
        sql_query = sql_query.filter(Challenge.transaction_id == transaction_id)

    if challenge is not None:
        sql_query = sql_query.filter(Challenge.challenge == challenge)

    challenges = sql_query.all()
    return challenges


@log_with(log)
def get_challenges_paginate(serial=None, transaction_id=None,
                            sortby=Challenge.timestamp,
                            sortdir="asc", psize=15, page=1):
    """
    This function is used to retrieve a challenge list, that can be displayed in
    the Web UI. It supports pagination.
    Each retrieved page will also contain a "next" and a "prev", indicating
    the next or previous page. If either does not exist, it is None.

    :param serial: The serial of the token
    :param transaction_id: The transaction_id of the challenge
    :param sortby: Sort by a Challenge DB field. The default is
        Challenge.timestamp.
    :type sortby: A Challenge column or a string.
    :param sortdir: Can be "asc" (default) or "desc"
    :type sortdir: basestring
    :param psize: The size of the page
    :type psize: int
    :param page: The number of the page to view. Starts with 1 ;-)
    :type page: int
    :return: dict with challenges, prev, next and count
    :rtype: dict
    """
    sql_query = _create_challenge_query(serial=serial,
                                        transaction_id=transaction_id)

    if isinstance(sortby, str):
        # convert the string to a Challenge column
        cols = Challenge.__table__.columns
        sortby = cols.get(sortby)

    if sortdir == "desc":
        sql_query = sql_query.order_by(sortby.desc())
    else:
        sql_query = sql_query.order_by(sortby.asc())

    pagination = db.paginate(sql_query, page=page, per_page=psize, error_out=False)
    challenges = pagination.items
    prev = None
    if pagination.has_prev:
        prev = page - 1
    next_page = None
    if pagination.has_next:
        next_page = page + 1
    challenge_list = []
    for challenge in challenges:
        challenge_dict = challenge.get()
        challenge_list.append(challenge_dict)

    ret = {"challenges": challenge_list,
           "prev": prev,
           "next": next_page,
           "current": page,
           "count": pagination.total}
    return ret


def _create_challenge_query(serial=None, transaction_id=None):
    """
    This function create the sql query for fetching transaction_ids. It is
    used by get_challenge_paginate.
    :return: An SQLAlchemy sql query
    """
    sql_query = Challenge.query
    if serial is not None and serial.strip("*"):
        # filter for serial
        if "*" in serial:
            # match with "like"
            sql_query = sql_query.filter(Challenge.serial.like(serial.replace(
                "*", "%")))
        else:
            # exact match
            sql_query = sql_query.filter(Challenge.serial == serial)

    if transaction_id is not None and transaction_id.strip("*"):
        # filter for serial
        if "*" in transaction_id:
            # match with "like"
            sql_query = sql_query.filter(Challenge.transaction_id.like(
                transaction_id.replace(
                    "*", "%")))
        else:
            # exact match
            sql_query = sql_query.filter(Challenge.transaction_id == transaction_id)

    return sql_query


def extract_answered_challenges(challenges):
    """
    Given a list of challenge objects, extract and return a list of *answered* challenge.
    A challenge is answered if it is not expired yet *and* if its ``otp_valid`` attribute
    is set to True.
    :param challenges: a list of challenge objects
    :return: a list of answered challenge objects
    """
    answered_challenges = []
    for challenge in challenges:
        # check if we are still in time.
        if challenge.is_valid():
            _, status = challenge.get_otp_status()
            if status is True:
                answered_challenges.append(challenge)
    return answered_challenges


def delete_challenges(serial: str = None, transaction_id: str = None) -> int:
    """
    This function deletes challenges from the database.

    :param serial: challenges for this very serial number
    :param transaction_id: challenges with this very transaction id
    :return: number of deleted challenges
    """
    challenges = get_challenges(serial=serial, transaction_id=transaction_id)
    for challenge in challenges:
        challenge.delete()
    return len(challenges)



#    E-mail: info@privacyidea.org
#    Contact: www.privacyidea.org
#
#    2018-06-15 Pascal Fuks <pascal@foxit.pro>
#               Added REGEXP parameter on phone number
#    2018-01-10 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#               Fix type cast for timeout
#    2016-06-14 Cornelius Kölbel <cornelius@privacyidea.org>
#               Add properties for new SMS provider model
#    2016-04-08 Cornelius Kölbel <cornelius@privacyidea.org>
#               Remote "None" as redundant 2nd argument to get
#    2016-01-13 Cornelius Kölbel <cornelius@privacyidea.org>
#               omit data object in GET request
#               omit params in POST request
#
#    privacyIDEA is a fork of LinOTP
#    May 28, 2014 Cornelius Kölbel
#    2015-01-30 Rewrite for migration to flask
#               Cornelius Kölbel <cornelius@privacyidea.org>
#
#
#    Copyright (C) LinOTP: 2010 - 2014 LSE Leading Security Experts GmbH
#
#    This program is free software: you can redistribute it and/or
#    modify it under the terms of the GNU Affero General Public
#    License, version 3, as published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the
#               GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#

__doc__ = """This is the SMSClass to send SMS via HTTP Gateways
It can handle HTTP/HTTPS POST and GET requests also with Proxy support

The code is tested in tests/test_lib_smsprovider
"""

from privacyidea.lib.smsprovider.SMSProvider import (ISMSProvider, SMSError)
from privacyidea.lib import _
import requests
import json
from urllib.parse import urlparse
import logging

log = logging.getLogger(__name__)


class HttpSMSProvider(ISMSProvider):

    def submit_message(self, phone, message):
        """
        send a message to a phone via an http sms gateway

        :param phone: the phone number
        :param message: the message to submit to the phone
        :return:
        """
        parameter = {}
        headers = {}
        if self.smsgateway:
            phone = self._mangle_phone(phone, self.smsgateway.option_dict)
            url = self.smsgateway.option_dict.get("URL")
            method = self.smsgateway.option_dict.get("HTTP_METHOD", "GET")
            username = self.smsgateway.option_dict.get("USERNAME")
            password = self.smsgateway.option_dict.get("PASSWORD")
            ssl_verify = self.smsgateway.option_dict.get("CHECK_SSL",
                                                         "yes") == "yes"
            json_data = self.smsgateway.option_dict.get("SEND_DATA_AS_JSON",
                                                        "no") == "yes"
            # FIXME: The Proxy option is deprecated and will be removed a version > 2.21
            proxy = self.smsgateway.option_dict.get("PROXY")
            http_proxy = self.smsgateway.option_dict.get('HTTP_PROXY')
            https_proxy = self.smsgateway.option_dict.get('HTTPS_PROXY')
            timeout = self.smsgateway.option_dict.get("TIMEOUT") or 3
            for k, v in self.smsgateway.option_dict.items():
                if k not in self.parameters().get("parameters"):
                    # This is an additional option
                    # We can not do .format() due to curly brackets in JSON
                    v = v.replace("{otp}", message)
                    v = v.replace("{phone}", phone)
                    try:
                        parameter[k] = json.loads(v)
                    except json.decoder.JSONDecodeError:
                        parameter[k] = v
            headers = self.smsgateway.header_dict
        else:
            phone = self._mangle_phone(phone, self.config)
            url = self.config.get('URL')
            method = self.config.get('HTTP_Method', 'GET')
            username = self.config.get('USERNAME')
            password = self.config.get('PASSWORD')
            ssl_verify = self.config.get('CHECK_SSL', True)
            json_data = False
            # FIXME: The Proxy option is deprecated and will be removed a version > 2.21
            proxy = self.config.get('PROXY')
            http_proxy = self.config.get('HTTP_PROXY')
            https_proxy = self.config.get('HTTPS_PROXY')
            parameter = self._get_parameters(message, phone)
            timeout = self.config.get("TIMEOUT") or 3

        log.debug("submitting message {0!r} to {1!s}".format(message, phone))

        if url is None:
            log.warning("can not submit message. URL is missing.")
            raise SMSError(-1, "No URL specified in the provider config.")
        basic_auth = None

        # there might be the basic authentication in the request url
        # like http://user:passw@hostname:port/path
        if password is None and username is None:
            parsed_url = urlparse(url)
            if "@" in parsed_url[1]:
                puser, server = parsed_url[1].split('@')
                username, password = puser.split(':')

        if username and password is not None:
            basic_auth = (username, password)

        proxies = {}
        if http_proxy:
            proxies["http"] = http_proxy
        if https_proxy:
            proxies["https"] = https_proxy
        if not proxies and proxy:
            # No new proxy config but only the old one.
            protocol = proxy.split(":")[0]
            proxies = {protocol: proxy}

        # url, parameter, username, password, method
        requestor = requests.get
        params = parameter
        data = None
        json_param = None
        if method == "POST":
            requestor = requests.post
            params = None
            if json_data:
                json_param = parameter
                log.debug("passing JSON data: {0!s}".format(json_param))
            else:
                data = parameter

        log_dict = {'params': params,
                    'headers': headers,
                    'method': method,
                    'basic_auth': basic_auth,
                    'url': url,
                    'data': data,
                    'json_param': json_param}
        log.debug("issuing request with parameters {params} (data: {data}, "
                  "json: {json_param}), headers {headers}, method {method} and"
                  "authentication {basic_auth} "
                  "to url {url}.".format(**log_dict))
        # Todo: drop basic auth if Authorization-Header is given?
        r = requestor(url, params=params, headers=headers,
                      data=data, json=json_param,
                      verify=ssl_verify,
                      auth=basic_auth,
                      timeout=float(timeout),
                      proxies=proxies)
        log.debug("queued SMS on the HTTP gateway. status code returned: {0!s}".format(
                  r.status_code))

        # We assume, that all gateways return with HTTP Status Code 200,
        # 201 or 202
        if r.status_code not in [200, 201, 202]:
            raise SMSError(r.status_code, "SMS could not be "
                                          "sent: %s" % r.status_code)
        success = self._check_success(r)
        return success

    def _get_parameters(self, message, phone):

        urldata = {}
        # transfer the phone key
        phoneKey = self.config.get('SMS_PHONENUMBER_KEY', "phone")
        urldata[phoneKey] = phone
        # transfer the sms key
        messageKey = self.config.get('SMS_TEXT_KEY', "sms")
        urldata[messageKey] = message
        params = self.config.get('PARAMETER', {})
        urldata.update(params)
        log.debug("[getParameters] urldata: {0!s}".format(urldata))
        return urldata

    def _check_success(self, response):
        """
        Check the success according to the reply
        1. if RETURN_SUCCESS is defined
        2. if RETURN_FAIL is defined
        :response reply: A response object.
        """
        reply = response.text
        ret = False
        if self.smsgateway:
            return_success = self.smsgateway.option_dict.get("RETURN_SUCCESS")
            return_fail = self.smsgateway.option_dict.get("RETURN_FAIL")
        else:
            return_success = self.config.get("RETURN_SUCCESS")
            return_fail = self.config.get("RETURN_FAIL")

        if return_success:
            if return_success in reply:
                log.debug("sending sms success")
                ret = True
            else:
                log.warning("failed to send sms. Reply %s does not match "
                            "the RETURN_SUCCESS definition" % reply)
                raise SMSError(response.status_code,
                               "We received a none success reply from the "
                               "SMS Gateway: {0!s} ({1!s})".format(reply,
                                                                   return_success))

        elif return_fail:
            if return_fail in reply:
                log.warning("sending sms failed. %s was not found "
                            "in %s" % (return_fail, reply))
                raise SMSError(response.status_code,
                               "We received the predefined error from the "
                               "SMS Gateway.")
            else:
                log.debug("sending sms success")
                ret = True
        else:
            ret = True
        return ret

    @classmethod
    def parameters(cls):
        """
        Return a dictionary, that describes the parameters and options for the
        SMS provider.
        Parameters are required keys to values.

        :return: dict
        """
        params = {"options_allowed": True,
                  "headers_allowed": True,
                  "parameters": {
                      "URL": {
                          "required": True,
                          "description": _("The base URL of the HTTP Gateway")},
                      "HTTP_METHOD": {
                          "required": True,
                          "description": _("Should the HTTP Gateway be "
                                           "connected via an HTTP GET or POST "
                                           "request."),
                          "values": ["GET", "POST"]},
                      "RETURN_SUCCESS": {
                          "description": _("Specify a substring, "
                                           "that indicates, that the SMS was "
                                           "delivered successfully.")},
                      "RETURN_FAIL": {
                          "description": _("Specify a substring, "
                                           "that indicates, that the SMS "
                                           "failed to be delivered.")},
                      "USERNAME": {
                          "description": _("Username in case of basic "
                                           "authentication.")
                      },
                      "PASSWORD": {
                          "description": _("Password in case of basic "
                                           "authentication.")
                      },
                      "CHECK_SSL": {
                          "required": True,
                          "description": _("Should the SSL certificate be "
                                           "verified."),
                          "values": ["yes", "no"]
                      },
                      "SEND_DATA_AS_JSON": {
                          "required": True,
                          "description": _("Should the data in a POST Request be sent "
                                           "as JSON."),
                          "values": ["yes", "no"]
                      },
                      "REGEXP": {
                          "description": cls.regexp_description
                      },
                      "PROXY": {"description": _("An optional proxy string. DEPRECATED. Do not use "
                                                 "this anymore. Rather use HTTP_PROXY for http "
                                                 "connections and HTTPS_PROXY for https "
                                                 "connection. The PROXY option will be removed in "
                                                 "future.")},
                      "HTTP_PROXY": {"description": _("Proxy setting for HTTP connections.")},
                      "HTTPS_PROXY": {"description": _("Proxy setting for HTTPS connections.")},
                      "TIMEOUT": {"description": _("The timeout in seconds.")}
                  }
                  }
        return params



# SPDX-FileCopyrightText: (C) 2024 Paul Lettich <paul.lettich@netknights.it>
#
# SPDX-FileCopyrightText: (C) 2017 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# Info: https://privacyidea.org
#
# This code is free software: you can redistribute it and/or
# modify it under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation, either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program. If not, see <http://www.gnu.org/licenses/>.
"""Utility functions for CLI tools"""
import click
from flask.cli import FlaskGroup
import platform
from privacyidea.app import create_app
from privacyidea.lib.utils import get_version_number


# Don't show logging information
def create_silent_app():
    """App factory with silent flag set"""
    app = create_app(config_name="production", silent=True)
    return app


# Don't load plugin commands
class NoPluginsFlaskGroup(FlaskGroup):
    """A FlaskGroup class which does not load commands from plugins"""

    def __init__(self, *args, **kwargs):
        # Hide the app option, we already hardcode the app in the CLI
        super().__init__(*args, **kwargs)
        for param in self.params:
            if param.name == "app":
                self.params.remove(param)

    def _load_plugin_commands(self):
        pass


def get_version(ctx, param, value):
    """Show version information"""
    if not value or ctx.resilient_parsing:
        return

    import werkzeug
    from flask import __version__

    message = "Python %(python)s\nFlask %(flask)s\nWerkzeug %(werkzeug)s\nprivacyIDEA %(privacyIDEA)s"
    click.echo(
        message
        % {
            "python": platform.python_version(),
            "flask": __version__,
            "werkzeug": werkzeug.__version__,
            "privacyIDEA": get_version_number(),
        },
        color=ctx.color,
    )
    ctx.exit()



# (c) NetKnights GmbH 2024,  https://netknights.it
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-FileCopyrightText: 2024 Nils Behlen <nils.behlen@netknights.it>
# SPDX-FileCopyrightText: 2024 Jelina Unger <jelina.unger@netknights.it>
# SPDX-License-Identifier: AGPL-3.0-or-later
#
import logging
from datetime import datetime, timezone

from typing import List, Union

from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePublicKey
from flask import json

from privacyidea.lib import _
from privacyidea.lib.challenge import get_challenges
from privacyidea.lib.config import get_token_types
from privacyidea.lib.containers.container_info import TokenContainerInfoData, PI_INTERNAL
from privacyidea.lib.crypto import verify_ecc, decryptPassword, FAILED_TO_DECRYPT_PASSWORD
from privacyidea.lib.error import ParameterError, ResourceNotFoundError, TokenAdminError
from privacyidea.lib.log import log_with
from privacyidea.lib.machine import is_offline_token
from privacyidea.lib.token import (create_tokenclass_object, get_tokens, get_serial_by_otp_list,
                                   get_tokens_from_serial_or_user)
from privacyidea.lib.tokenclass import TokenClass
from privacyidea.lib.user import User
from privacyidea.lib.utils import is_true
from privacyidea.models import (TokenContainerOwner, Realm, Token, db, TokenContainerStates,
                                TokenContainerInfo, TokenContainerRealm, TokenContainerTemplate)

log = logging.getLogger(__name__)


class TokenContainerClass:
    options = {}

    @log_with(log)
    def __init__(self, db_container):
        self._db_container = db_container
        # Create the TokenClass objects from the database objects
        token_list = []
        for t in db_container.tokens:
            token_object = create_tokenclass_object(t)
            if isinstance(token_object, TokenClass):
                token_list.append(token_object)

        self.tokens = token_list
        self.options = {}

    @classmethod
    def get_class_options(cls, only_selectable=False) -> dict:
        """
        Returns the options for the container class.

        :param only_selectable: If True, only options with more than one value are returned.
        :return: Dictionary in the format {key: [values]}
        """
        if only_selectable:
            class_options = {key: values for key, values in cls.options.items() if len(values) > 1}
        else:
            class_options = cls.options
        return class_options

    def set_default_option(self, key) -> str:
        """
        Checks if a value is set in the container info for the requested key.
        If not, the default value is set, otherwise the already set value is kept.

        :param key: The key to be checked.
        :return: The used value for the key or None if the key does not exist in the class options.
        """
        class_options = self.get_class_options()
        key_options = class_options.get(key)

        if key_options is None:
            # The key does not exist in the class options
            return None

        # Check if value is already set for this key in the container info
        container_info = self.get_container_info_dict()
        value = container_info.get(key)
        if not value:
            # key not defined: set default value
            value = class_options[key][0]
            self.update_container_info([TokenContainerInfoData(key=key, value=value, info_type=PI_INTERNAL)])
        return value

    @property
    def serial(self) -> str:
        return self._db_container.serial

    @property
    def description(self) -> str:
        return self._db_container.description

    @description.setter
    def description(self, value: str):
        if not value:
            value = ""
        self._db_container.description = value
        self._db_container.save()

    @property
    def type(self) -> str:
        return self._db_container.type

    @property
    def db_id(self) -> int:
        return self._db_container.id

    @property
    def last_authentication(self) -> Union[datetime, None]:
        """
        Returns the timestamp of the last seen field in the database.
        It is the time when a token of the container was last used successfully for authentication.
        From this layer on it is hence called 'last_authentication'.
        If the container was never used for authentication, the value is None.
        """
        last_auth = self._db_container.last_seen
        if last_auth:
            last_auth = last_auth.replace(tzinfo=timezone.utc)
        return last_auth

    def update_last_authentication(self):
        """
        Updates the timestamp of the last seen field in the database.
        """
        # SQLite does not support timezone aware timestamps, hence all time stamps are stored in utc time.
        # The timezone information must be removed, because some databases would change the time stamp to local time
        # (e.g. postgresql)
        self._db_container.last_seen = datetime.now(timezone.utc).replace(tzinfo=None)
        self._db_container.save()

    def reset_last_authentication(self):
        """
        Resets the timestamp of the last seen field in the database.
        """
        self._db_container.last_seen = None
        self._db_container.save()

    @property
    def last_synchronization(self) -> Union[datetime, None]:
        """
        Returns the timestamp of the last updated field in the database.
        It is the time when the container was last synchronized with the privacyIDEA server.
        From this layer on it is hence called 'last_synchronization'.
        """
        last_sync = self._db_container.last_updated
        if last_sync:
            last_sync = last_sync.replace(tzinfo=timezone.utc)
        return last_sync

    def update_last_synchronization(self):
        """
        Updates the timestamp of the last updated field in the database.
        """
        # SQLite does not support timezone aware timestamps, hence all time stamps are stored in utc time.
        # The timezone information must be removed, because some databases would change the time stamp to local time
        # (e.g. postgresql)
        self._db_container.last_updated = datetime.now(timezone.utc).replace(tzinfo=None)
        self._db_container.save()

    def reset_last_synchronization(self):
        """
        Resets the timestamp of the last updated field in the database.
        """
        self._db_container.last_updated = None
        self._db_container.save()

    @property
    def realms(self) -> list:
        return self._db_container.realms

    def set_realms(self, realms, add=False) -> dict:
        """
        Set the realms of the container. If `add` is True, the realms will be added to the existing realms, otherwise
        the existing realms will be removed.

        :param realms: List of realm names
        :param add: False if the existing realms shall be removed, True otherwise
        :return: Dictionary in the format {realm: success}, the entry 'deleted' indicates whether existing realms were
                 deleted.
        """
        result = {}

        if not realms:
            realms = []

        # delete all container realms
        if not add:
            TokenContainerRealm.query.filter_by(container_id=self._db_container.id).delete()
            result["deleted"] = True
            self._db_container.save()

            # Check that user realms are kept
            user_realms = self._get_user_realms()
            missing_realms = list(set(user_realms).difference(realms))
            realms.extend(missing_realms)
            for realm in missing_realms:
                log.warning(
                    f"Realm {realm} can not be removed from container {self.serial} "
                    f"because a user from this realm is assigned th the container.")
        else:
            result["deleted"] = False

        for realm in realms:
            if realm:
                realm_db = Realm.query.filter_by(name=realm).first()
                if not realm_db:
                    result[realm] = False
                    log.warning(f"Realm {realm} does not exist.")
                else:
                    realm_id = realm_db.id
                    # Check if realm is already assigned to the container
                    if not TokenContainerRealm.query.filter_by(container_id=self._db_container.id,
                                                               realm_id=realm_id).first():
                        self._db_container.realms.append(realm_db)
                        result[realm] = True
                    else:
                        log.info(f"Realm {realm} is already assigned to container {self.serial}.")
                        result[realm] = False
        self._db_container.save()

        return result

    def _get_user_realms(self) -> list:
        """
        Returns a list of the realm names of the users that are assigned to the container.
        """
        owners = self.get_users()
        realms = [owner.realm for owner in owners]
        return realms

    def remove_token(self, serial: str) -> bool:
        """
        Remove a token from the container. Raises a ResourceNotFoundError if the token does not exist.

        :param serial: Serial of the token
        :return: True if the token was successfully removed, False if the token was not found in the container
        """
        token = Token.query.filter(Token.serial == serial).first()
        if not token:
            raise ResourceNotFoundError(f"Token with serial {serial} does not exist.")
        if token not in self._db_container.tokens:
            log.info(f"Token with serial {serial} not found in container {self.serial}.")
            return False

        self._db_container.tokens.remove(token)
        self._db_container.save()
        self.tokens = [t for t in self.tokens if t.get_serial() != serial]
        return True

    def add_token(self, token: TokenClass) -> bool:
        """
        Add a token to the container.
        Raises a ParameterError if the token type is not supported by the container.

        :param token: TokenClass object
        :return: True if the token was successfully added, False if the token is already in the container
        """
        if token.get_type() not in self.get_supported_token_types():
            raise ParameterError(f"Token type {token.get_type()} not supported for container type {self.type}. "
                                 f"Supported types are {self.get_supported_token_types()}.")
        if token.get_serial() not in [t.get_serial() for t in self.tokens]:
            self.tokens.append(token)
            self._db_container.tokens = [t.token for t in self.tokens]
            self._db_container.save()
            return True
        return False

    def get_tokens(self) -> list:
        """
        Returns the tokens of the container as a list of TokenClass objects.
        """
        return self.tokens

    def get_tokens_for_synchronization(self) -> list:
        """
        Returns the tokens of the container that can be synchronized with a client as a list of TokenClass objects.
        """
        return self.tokens

    def delete(self) -> int:
        """
        Deletes the container and all associated objects from the database.
        Returns the id of the deleted container.
        """
        return self._db_container.delete()

    def add_user(self, user: User) -> bool:
        """
        Assign a user to the container.
        Raises a UserError if the user does not exist.
        Raises a TokenAdminError if the container already has an owner.

        :param user: User object
        :return: True if the user was assigned
        """
        (user_id, resolver_type, resolver_name) = user.get_user_identifiers()
        if not self._db_container.owners.first():
            TokenContainerOwner(container_id=self._db_container.id,
                                user_id=user_id,
                                resolver=resolver_name,
                                realm_id=user.realm_id).save()
            # Add user realm to container realms
            realm_db = Realm.query.filter_by(name=user.realm).first()
            self._db_container.realms.append(realm_db)
            self._db_container.save()
            return True
        log.info(f"Container {self.serial} already has an owner.")
        raise TokenAdminError("This container is already assigned to another user.")

    def remove_user(self, user: User) -> bool:
        """
        Remove a user from the container. Raises a ResourceNotFoundError if the user does not exist.

        :param user: User object to be removed
        :return: True if the user was removed, False if the user was not found in the container
        """
        (user_id, resolver_type, resolver_name) = user.get_user_identifiers()
        count = TokenContainerOwner.query.filter_by(container_id=self._db_container.id,
                                                    user_id=user_id,
                                                    resolver=resolver_name).delete()
        db.session.commit()
        return count > 0

    def get_users(self) -> list:
        """
        Returns a list of users that are assigned to the container.
        """
        db_container_owners: List[TokenContainerOwner] = TokenContainerOwner.query.filter_by(
            container_id=self._db_container.id).all()

        users: list[User] = []
        for owner in db_container_owners:
            realm = Realm.query.filter_by(id=owner.realm_id).first()
            user = User(uid=owner.user_id, realm=realm.name, resolver=owner.resolver)
            users.append(user)

        return users

    def get_states(self) -> list:
        """
        Returns the states of the container as a list of strings.
        """
        db_states = self._db_container.states
        states = [state.state for state in db_states]
        return states

    def _check_excluded_states(self, states) -> bool:
        """
        Validates whether the state list contains states that excludes each other

        :param states: list of states
        :returns: True if the state list contains exclusive states, False otherwise
        """
        state_types = self.get_state_types()
        for state in states:
            if state in state_types:
                excluded_states = state_types[state]
                same_states = list(set(states).intersection(excluded_states))
                if len(same_states) > 0:
                    return True
        return False

    def set_states(self, state_list: list) -> dict:
        """
        Set the states of the container. Previous states will be removed.
        Raises a ParameterError if the state list contains exclusive states.

        :param state_list: List of states as strings
        :returns: Dictionary in the format {state: success}
        """
        if not state_list:
            state_list = []

        # Check for exclusive states
        exclusive_states = self._check_excluded_states(state_list)
        if exclusive_states:
            raise ParameterError(f"The state list {state_list} contains exclusive states!")

        # Remove old state entries
        TokenContainerStates.query.filter_by(container_id=self._db_container.id).delete()

        # Set new states
        state_types = self.get_state_types().keys()
        res = {}
        for state in state_list:
            if state not in state_types:
                # We do not raise an error here to allow following states to be set
                log.warning(f"State {state} not supported. Supported states are {state_types}.")
                res[state] = False
            else:
                TokenContainerStates(container_id=self._db_container.id, state=state).save()
                res[state] = True
        return res

    def add_states(self, state_list: list) -> dict:
        """
        Add states to the container. Previous states are only removed if a new state excludes them.
        Raises a ParameterError if the state list contains exclusive states.

        :param state_list: List of states as strings
        :returns: Dictionary in the format {state: success}
        """
        if not state_list or len(state_list) == 0:
            return {}

        # Check for exclusive states
        exclusive_states = self._check_excluded_states(state_list)
        if exclusive_states:
            raise ParameterError(f"The state list {state_list} contains exclusive states!")

        # Add new states
        res = {}
        state_types = self.get_state_types()
        for state in state_list:
            if state not in state_types.keys():
                # We do not raise an error here to allow following states to be set
                res[state] = False
                log.warning(f"State {state} not supported. Supported states are {state_types}.")
            else:
                # Remove old states that are excluded from the new state
                for excluded_state in state_types[state]:
                    TokenContainerStates.query.filter_by(container_id=self._db_container.id,
                                                         state=excluded_state).delete()
                    log.debug(
                        f"Removed state {excluded_state} from container {self.serial} "
                        f"because it is excluded by the new state {state}.")
                TokenContainerStates(container_id=self._db_container.id, state=state).save()
                res[state] = True
        return res

    @classmethod
    def get_state_types(cls) -> dict:
        """
        Returns the state types that are supported by this container class and the states that are exclusive
        to each of these states.

        :return: Dictionary in the format: {state: [excluded_states]}
        """
        state_types_exclusions = {
            "active": ["disabled"],
            "disabled": ["active"],
            "lost": [],
            "damaged": []
        }
        return state_types_exclusions

    def set_container_info(self, info):
        """
        Set the containerinfo field in the DB. Old values will be deleted.

        :param info: dictionary in the format: {key: value}
        """
        self.delete_container_info(keep_internal=True)
        if info:
            self._db_container.set_info(info)

    def update_container_info(self, info: list):
        """
        Updates the container info for the passed list of container info. Non-existing keys are added and the values
        for existing keys are updated.

        :param info: list of TokenContainerInfoData objects
        """
        for data in info:
            TokenContainerInfo(self.db_id, data.key, data.value, type=data.type, description=data.description).save(
                persistent=False)
        db.session.commit()

    def get_container_info(self) -> list:
        """
        Return the tokencontainerinfo from the DB

        :return: list of tokencontainerinfo objects
        """
        return self._db_container.info_list

    def get_internal_info_keys(self) -> list:
        """
        Returns the keys of the internal container info.
        """
        return [info.key for info in self.get_container_info() if info.type == PI_INTERNAL]

    def get_container_info_dict(self) -> dict:
        """
        Return the tokencontainerinfo from the DB as dictionary

        :return: dictionary of tokencontainerinfo objects
        """
        container_info_list = self._db_container.info_list
        container_info_dict = {info.key: info.value for info in container_info_list}
        return container_info_dict

    def delete_container_info(self, key=None, keep_internal: bool = True) -> dict:
        """
        Delete the tokencontainerinfo from the DB

        :param key: key to delete, if None all keys are deleted
        :param keep_internal: If True, entries of type PI_INTERNAL are not deleted
        :return: dictionary of deleted keys in the format {key: deleted}
        """
        res = {}
        if key:
            container_infos = TokenContainerInfo.query.filter_by(container_id=self._db_container.id, key=key)
        else:
            container_infos = TokenContainerInfo.query.filter_by(container_id=self._db_container.id)

        if container_infos.count() == 0:
            res[key] = False
            log.debug(f"Container {self.serial} has no info with key {key} or no info at all.")

        for ci in container_infos:
            if not keep_internal or ci.type != PI_INTERNAL:
                ci.delete()
                res[ci.key] = True
            else:
                log.debug(f"Container info with key {ci.key} is of type {PI_INTERNAL} and can not be deleted.")
                res[ci.key] = False
        return res

    def add_options(self, options):
        """
        Add the given options to the container.

        :param options: The options to add as dictionary
        """
        class_options = self.get_class_options()
        for key, value in options.items():
            option_values = class_options.get(key)
            if option_values is not None:
                if value in option_values:
                    self.update_container_info([TokenContainerInfoData(key=key, value=value, info_type=PI_INTERNAL)])
                else:
                    log.debug(f"Value {value} not supported for option key {key}.")
            else:
                log.debug(f"Option key {key} not found for container type {self.get_class_type()}.")

    @property
    def template(self) -> TokenContainerTemplate:
        """
        Returns the template the container is based on.
        """
        return self._db_container.template

    @template.setter
    def template(self, template_name: str):
        """
        Set the template the container is based on.
        """
        db_template = TokenContainerTemplate.query.filter_by(name=template_name).first()
        if db_template:
            if db_template.container_type == self.type:
                self._db_container.template = db_template
                self._db_container.save()
            else:
                log.info(f"Template {template_name} is not compatible with container type {self.type}.")

    def init_registration(self, server_url: str, scope: str, registration_ttl: int, ssl_verify: bool,
                          params: dict = None):
        """
        Initializes the registration: Generates a QR code containing all relevant data.

        :param server_url: URL of the server reachable for the client.
        :param scope: The URL the client contacts to finalize the registration e.g. "https://pi.net/container/register/finalize".
        :param registration_ttl: Time to live of the registration link in minutes.
        :param ssl_verify: Whether the client shall use ssl.
        :param params: Container specific parameters
        """
        raise NotImplementedError("Registration is not implemented for this container type.")

    def finalize_registration(self, params):
        """
        Finalize the registration of a container.
        """
        raise NotImplementedError("Registration is not implemented for this container type.")

    def terminate_registration(self):
        """
        Terminate the synchronisation of the container with privacyIDEA.
        """
        raise NotImplementedError("Registration is not implemented for this container type.")

    def check_challenge_response(self, params: dict):
        """
        Checks if the response to a challenge is valid.
        """
        return False

    def create_challenge(self, scope, validity_time=2):
        """
        Create a challenge for the container.
        """
        return {}

    def validate_challenge(self, signature: bytes, public_key: EllipticCurvePublicKey, scope: str,
                           transaction_id: str = None, key: str = None, container: str = None, device_brand: str = None,
                           device_model: str = None) -> bool:
        """
        Verifies the response of a challenge:
            * Checks if challenge is valid (not expired)
            * Checks if the challenge is for the right scope
            * Verifies the signature
        Implicitly verifies the passphrase by adding it to the signature message. The passphrase needs to be defined in
        the challenge data. Otherwise, no passphrase is used.

        :param signature: Signature of the message
        :param public_key: Public key to verify the signature
        :param scope: endpoint to reach if the challenge is valid
        :param transaction_id: Transaction ID of the challenge, optional
        :param key: Key to be included in the signature, optional
        :param container: Container to be included in the signature, optional
        :param device_brand: Device brand to be included in the signature, optional
        :param device_model: Device model to be included in the signature, optional
        :return: True if the challenge response is valid, False otherwise
        """
        challenge_list = get_challenges(serial=self.serial, transaction_id=transaction_id)
        container_info = self.get_container_info_dict()
        hash_algorithm = container_info.get("hash_algorithm", "SHA256")
        verify_res = {"valid": False, "hash_algorithm": hash_algorithm}

        # Checks all challenges of the container, at least one must be valid
        for challenge in challenge_list:
            if challenge.is_valid():
                # Create message
                nonce = challenge.challenge
                times_stamp = challenge.timestamp.replace(tzinfo=timezone.utc).isoformat()
                extra_data = json.loads(challenge.data)
                passphrase = extra_data.get("passphrase_response")
                challenge_scope = extra_data.get("scope")
                # explicitly check the scope that the right challenge for the right endpoint is used
                if scope != challenge_scope:
                    log.debug(f"Scope {scope} does not match challenge scope {challenge_scope}.")
                    continue

                message = f"{nonce}|{times_stamp}|{self.serial}|{challenge_scope}"
                if device_brand:
                    message += f"|{device_brand}"
                if device_model:
                    message += f"|{device_model}"
                if passphrase:
                    passphrase = decryptPassword(passphrase)
                    if passphrase == FAILED_TO_DECRYPT_PASSWORD:
                        challenge.delete()
                        log.warning("Failed to decrypt the passphrase from the challenge. "
                                    "Hence, deleted the challenge.")
                        continue
                    else:
                        message += f"|{passphrase}"
                if key:
                    message += f"|{key}"
                if container:
                    message += f"|{container}"

                # Check signature
                try:
                    verify_res = verify_ecc(message.encode("utf-8"), signature, public_key, hash_algorithm)
                except InvalidSignature:
                    # It is not the right challenge: log to find reason for invalid signature
                    log.debug(f"Used hash algorithm to verify: {verify_res['hash_algorithm']}")
                    challenge_data_log = (f"Challenge data: nonce={nonce}, timestamp={times_stamp}, "
                                          f"serial={self.serial}, scope={scope}")
                    if passphrase:
                        challenge_data_log += f", passphrase={len(passphrase) * '*'}"
                    log.debug(challenge_data_log)
                    log.debug(f"Challenge data from the client: device_brand={device_brand}, "
                              f"device_model={device_model}, key={key}, container={container} ")
                    continue

                # Valid challenge: delete it
                challenge.delete()
                break
            else:
                # Delete expired challenge
                challenge.delete()

        return verify_res["valid"]

    def get_as_dict(self, include_tokens: bool = True, public_info: bool = True,
                    additional_hide_info: list = None) -> dict:
        """
        Returns a dictionary containing all properties, contained tokens, and owners

        :param include_tokens: If True, the tokens are included in the dictionary
        :param public_info: If True, only public information is included and sensitive information is omitted
        :param additional_hide_info: List of keys that shall be omitted from the dictionary
        :return: Dictionary with the container details

        Example response

        ::

            {
                "type": "smartphone",
                "serial": "SMPH00038DD3",
                "description": "My smartphone",
                "last_authentication": "2024-09-11T08:56:37.200336+00:00",
                "last_synchronization": "2024-09-11T08:56:37.200336+00:00",
                "states": ["active"],
                "info": {
                            "hash_algorithm": "SHA256",
                            "key_algorithm": "secp384r1"
                        },
                "internal_info_keys": ["hash_algorithm"],
                "realms": ["deflocal"],
                "users": [{"user_name": "testuser",
                           "user_realm": "deflocal",
                           "user_resolver": "internal",
                           "user_id": 1}],
                "tokens": ["TOTP000152D1", "HOTP00012345"]
            }
        """
        details = {"type": self.type,
                   "serial": self.serial,
                   "description": self.description,
                   "last_authentication": self.last_authentication.isoformat() if self.last_authentication else None,
                   "last_synchronization": self.last_synchronization.isoformat() if self.last_synchronization else None,
                   "states": self.get_states()}

        if public_info or additional_hide_info:
            key_blacklist = []
            if public_info:
                key_blacklist = ["public_key_client", "rollover_server_url", "rollover_challenge_ttl"]
            if additional_hide_info:
                key_blacklist.extend(additional_hide_info)
            info = self.get_container_info()
            info_dict = {i.key: i.value for i in info if i.key not in key_blacklist}
        else:
            info_dict = self.get_container_info_dict()
        details["info"] = info_dict
        details["internal_info_keys"] = self.get_internal_info_keys()

        template = self.template
        template_name = ""
        if template:
            template_name = template.name
        details["template"] = template_name

        realms = []
        for realm in self.realms:
            realms.append(realm.name)
        details["realms"] = realms

        users = []
        user_info = {}
        for user in self.get_users():
            user_info["user_name"] = user.login
            user_info["user_realm"] = user.realm
            user_info["user_resolver"] = user.resolver
            user_info["user_id"] = user.uid
            users.append(user_info)
        details["users"] = users

        if include_tokens:
            details["tokens"] = [token.get_serial() for token in self.get_tokens()]

        return details

    @classmethod
    def get_class_type(cls) -> str:
        """
        Returns the type of the container class.
        """
        return "generic"

    @classmethod
    def get_supported_token_types(cls) -> list:
        """
        Returns the token types that are supported by the container class.
        """
        supported_token_types = get_token_types()
        supported_token_types.sort()
        return supported_token_types

    @classmethod
    def get_class_prefix(cls) -> str:
        """
        Returns the container class specific prefix for the serial.
        """
        return "CONT"

    @classmethod
    def get_class_description(cls) -> str:
        """
        Returns a description of the container class.
        """
        return _("General purpose container that can hold any type and any number of token.")

    def encrypt_dict(self, container_dict: dict, params: dict):
        """
        Encrypts a dictionary with the public key of the container.
        It is not supported by all container classes. Classes not supporting the encryption raise a privacyIDEA error.
        """
        raise NotImplementedError("Encryption is not implemented for this container type.")

    def synchronize_container_details(self, container_client: dict,
                                      initial_transfer_allowed: bool = False) -> dict:
        """
        Compares the container from the client with the server and returns the differences.
        The container dictionary from the client contains information about the container itself and the tokens.
        For each token the type and serial shall be provided. If no serial is available, two otp values can be provided.
        The server than tries to find the serial for the otp values. If multiple serials are found, it will not be
        included in the returned dictionary, since the token can not be uniquely identified.
        The returned dictionary contains information about the container itself and the tokens that needs to be added
        or updated. For the tokens to be added the enrollUrl is provided. For the tokens to be updated at least the
        serial and the tokentype are provided.

        :param initial_transfer_allowed: If True, all tokens from the client are added to the container
        :param container_client: The container from the client as dictionary.
        An example container dictionary from the client:
            ::

                {
                    "serial": "SMPH001",
                    "type": "smartphone",
                    "tokens": [{"serial": "TOTP001", "tokentype": "TOTP"},
                                {"otp": ["1234", "9876"], "tokentype": "HOTP", "counter": "2"}]
                }

        :return: container dictionary
        An example of a returned container dictionary:
            ::

                {
                    "container": {"type": "smartphone", "serial": "SMPH001"},
                    "tokens": {"add": ["enroll_url1", "enroll_url2"],
                               "update": [{"serial": "TOTP001", "tokentype": "totp"},
                                          {"serial": "HOTP001", "otp": ["1234", "9876"],
                                           "tokentype": "hotp", "counter": 2}]}
                }
        """
        container_dict = {"container": {"type": self.type, "serial": self.serial}}
        server_token_serials = [token.get_serial() for token in self.get_tokens_for_synchronization()]

        # Get serials for client tokens without serial
        client_tokens = container_client.get("tokens", [])
        serial_otp_map = {}
        for token in client_tokens:
            dict_keys = token.keys()
            # Get serial from otp if required
            if "serial" not in dict_keys and "otp" in dict_keys:
                token_type = token.get("tokentype")
                if token_type:
                    token_type = token_type.lower()
                token_list = get_tokens(tokentype=token_type)
                client_counter = token.get("counter")
                if client_counter:
                    client_counter = int(client_counter)
                serial_list = get_serial_by_otp_list(token_list, otp_list=token["otp"], counter=client_counter)
                if len(serial_list) == 1:
                    serial = serial_list[0]
                    token["serial"] = serial
                    serial_otp_map[serial] = token["otp"]
                elif len(serial_list) > 1:
                    log.warning(f"Multiple serials found for otp {token['otp']}. Ignoring this token.")
                # shall we ignore otp values where multiple serials are found?

        # map client and server tokens
        client_serials = [token["serial"] for token in client_tokens if "serial" in token.keys()]

        container_info = self.get_container_info_dict()
        registration_state = container_info.get("registration_state", "")
        if registration_state == "rollover_completed":
            # rollover all tokens: generate new enroll info for all tokens
            missing_serials = server_token_serials
            same_serials = []
        else:
            missing_serials = list(set(server_token_serials).difference(set(client_serials)))
            same_serials = list(set(server_token_serials).intersection(set(client_serials)))

        # Offline tokens that do not exist on the client can not be added during synchronization since the offline
        # counter is not known by the server
        offline_serials = [serial for serial in missing_serials if is_offline_token(serial)]
        missing_serials = list(set(missing_serials).difference(set(offline_serials)))
        if registration_state == "rollover_completed":
            # offline tokens known by the client can be added to the update list
            # (indicating the tokens are still in the container)
            same_serials = list(set(client_serials).intersection(set(offline_serials)))
            # only offline tokens not known to the client are excluded (they can not be added to the client)
            offline_serials = list(set(offline_serials).difference(set(client_serials)))
        if len(offline_serials) > 0:
            log.info(f"The following offline tokens do not exist on the client: {offline_serials}. "
                     "They can not be added during synchronization.")

        # Initial synchronization after registration or rollover
        if initial_transfer_allowed and not is_true(container_info.get("initially_synchronized")):
            self.update_container_info(
                [TokenContainerInfoData(key="initially_synchronized", value="True", info_type=PI_INTERNAL)])
            server_missing_tokens = list(set(client_serials).difference(set(server_token_serials)))
            for serial in server_missing_tokens:
                # Try to add the missing token to the container on the server
                try:
                    token = get_tokens_from_serial_or_user(serial, None)[0]
                except ResourceNotFoundError:
                    log.info(f"Token {serial} from client does not exist on the server.")
                    continue

                try:
                    self.add_token(token)
                except ParameterError as e:
                    log.info(f"Client token {serial} could not be added to the container: {e}")
                    continue
                # add token to the same_serials list to update the token details
                same_serials.append(serial)

        # Get info for same serials: token details
        update_dict = []
        for serial in same_serials:
            token = get_tokens_from_serial_or_user(serial, None)[0]
            offline = is_offline_token(serial)
            token_dict_all = token.get_as_dict()
            token_dict = {"serial": token_dict_all["serial"], "tokentype": token_dict_all["tokentype"],
                          "offline": offline}
            # rename count to counter for the client
            if "count" in token_dict_all:
                # For offline tokens the counter shall not be sent to the client
                if not offline:
                    token_dict["counter"] = token_dict_all["count"]

            # add otp values to allow the client identifying the token if he has no serial yet
            otp = serial_otp_map.get(serial)
            if otp:
                token_dict["otp"] = otp
            update_dict.append(token_dict)

        container_dict["tokens"] = {"add": missing_serials, "update": update_dict}

        if registration_state == "rollover_completed":
            container_dict["tokens"]["offline"] = offline_serials

        return container_dict



"""
This test file tests the lib.serviceid methods.

This tests the token functions on an interface level
"""
from .base import MyTestCase
from privacyidea.lib.error import privacyIDEAError, ResourceNotFoundError

from privacyidea.lib.serviceid import set_serviceid, delete_serviceid, get_serviceids
from privacyidea.models import Serviceid

WEBSERVER = "webserver"
MAILSERVER = "mailserver"


class TokenTestCase(MyTestCase):

    def test_01_create_serviceid(self):

        r = set_serviceid(WEBSERVER, "all cool light machines")
        self.assertGreaterEqual(r, 1)
        si = Serviceid.query.filter_by(id=r).first()
        self.assertEqual(si.Description, "all cool light machines")

        # Update
        r = set_serviceid(WEBSERVER, "all heavy httpd machines")
        self.assertGreaterEqual(r, 1)
        si = Serviceid.query.filter_by(id=r).first()
        self.assertEqual(si.Description, "all heavy httpd machines")

    def test_02_delete_serviceid(self):
        r = set_serviceid(WEBSERVER, "HTTP")
        self.assertGreaterEqual(r, 1)

        delete_serviceid(WEBSERVER)
        si = Serviceid.query.filter_by(name=WEBSERVER).all()
        self.assertEqual(len(si), 0)

        r = set_serviceid("webserver", "other machines")
        self.assertGreaterEqual(r, 1)

        self.assertRaises(privacyIDEAError,
                          delete_serviceid, sid=(r + 1), name=WEBSERVER)

        delete_serviceid(sid=r)
        si = Serviceid.query.filter_by(name=WEBSERVER).all()
        self.assertEqual(len(si), 0)

        self.assertRaises(ResourceNotFoundError, delete_serviceid)

    def test_03_get_serviceids(self):
        r1 = set_serviceid(WEBSERVER, "httpd")
        self.assertGreaterEqual(r1, 1)

        r2 = set_serviceid(MAILSERVER, "smtpd")
        self.assertGreater(r2, r1)

        sis = get_serviceids()
        self.assertEqual(len(sis), 2)

        sis = get_serviceids(name=WEBSERVER)
        self.assertEqual(len(sis), 1)

        sis = get_serviceids(id=r2)
        self.assertEqual(len(sis), 1)
        self.assertEqual(sis[0].name, MAILSERVER)



#!/usr/bin/python
#  2018-05-27 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             init

__doc__ = """
You can use this script to update counter values of tokens in the privacyIDEA
database. This is helpful after database migrations. Counters in the old
instance may have been updated, since users still authenticate against this
old instance. You would simply fetch these updated counters and use this
script to update them in the new privacyIDEA database.

You can update counters like

privacyidea-export-linotp-counter.py -c MIGRATION/linotp.ini  | ./tools/privacyidea-update-counter.py -c /etc/privacyidea/pi.cfg -i -
"""
from sqlalchemy.orm import sessionmaker

import argparse
import sys
from sqlalchemy import create_engine
from sqlalchemy import MetaData, Column
from sqlalchemy import Integer, Unicode, Boolean
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()
metadata = MetaData()


class LinToken(Base):
    __tablename__ = "Token"

    LinOtpTokenId = Column(Integer, primary_key=True, nullable=False)
    LinOtpTokenDesc = Column(Unicode(80), default='')
    LinOtpTokenSerialnumber = Column(Unicode(40), default='', unique=True, nullable=False, index=True)
    LinOtpTokenType = Column(Unicode(30), default='HMAC',index=True)
    LinOtpTokenInfo = Column(Unicode(2000), default='')
    LinOtpTokenPinUser = Column(Unicode(512), default='')
    LinOtpTokenPinUserIV =Column(Unicode(32), default='')
    LinOtpTokenPinSO = Column(Unicode(512), default='')
    LinOtpTokenPinSOIV = Column(Unicode(32), default='')
    LinOtpIdResolver = Column(Unicode(120), default='', index=True)
    LinOtpIdResClass = Column(Unicode(120), default='')
    LinOtpUserid = Column(Unicode(320), default='', index=True)
    LinOtpSeed = Column(Unicode(32), default='')
    LinOtpOtpLen = Column(Integer, default=6)
    LinOtpPinHash = Column(Unicode(512), default='')
    LinOtpKeyEnc = Column(Unicode(1024), default='')
    LinOtpKeyIV = Column(Unicode(32), default='')
    LinOtpMaxFail = Column(Integer, default=10)
    LinOtpIsactive = Column(Boolean(), default=True)
    LinOtpFailCount = Column(Integer, default=0)
    LinOtpCount = Column(Integer, default=0)
    LinOtpCountWindow = Column(Integer, default=10)
    LinOtpSyncWindow = Column(Integer, default=1000)


def get_linotp_uri(config_file):
    with open(config_file) as f:
        content = f.readlines()

    lines = [l.strip() for l in content]
    sql_uri = ""
    for line in lines:
        if line.startswith("SQLALCHEMY_DATABASE_URI"):
            sql_uri = line.split("=", 1)[1].strip().strip("'").strip('"')
    return sql_uri


def read_counter_file(import_file):
    update_list = []
    for line in import_file.readlines():
        try:
            serial, counter = [v.strip() for v in line.split(",")]
            update_list.append(("{0!s}".format(serial), int(counter)))
        except ValueError as ve:
            # If there is a line, that does not comply
            sys.stderr.write("Failed to parse line: {0!s}\n".format(line))

    return update_list


def main():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("-c", "--config", help="privacyIDEA config file. We only need the SQLALCHEMY_DATABASE_URI.",
                        required=True)
    parser.add_argument('file', help='The CSV file with the updated counters. The file should contain one serial and '
                                     'counter per line split by a comma. You can specify "-" to read from stdin.',
                        type=argparse.FileType())
    parser.add_argument("-i", "--increase-only", help="Only update the token counter, if the new counter value"
                                                      "is bigger than the existing in the database.",
                        action='store_const', const=True)
    args = parser.parse_args()

    # Parse data

    SQL_URI = get_linotp_uri(args.config)
    counters = read_counter_file(args.file)

    # Start DB stuff

    linotp_engine = create_engine(SQL_URI)
    linotp_session = sessionmaker(bind=linotp_engine)()

    print("Starting updating {0!s} counters:".format(len(counters)))
    updated = 0
    not_found = 0
    processed = 0
    unknown_tokens = []
    for count in counters:
        processed += 1
        if args.increase_only:
            r = linotp_session.query(LinToken).filter_by(LinOtpTokenSerialnumber=count[0]).first()
            if r and r.LinOtpCount >= count[1]:
                # The counter in the database is bigger
                continue
        sys.stdout.write("\r {0!s}: {1!s}     ".format(processed, count[0]))
        r = linotp_session.query(LinToken).filter_by(LinOtpTokenSerialnumber=count[0]).update({"LinOtpCount": count[1]})
        if r > 0:
            # r==0, if the token was not found!
            updated += 1
        else:
            not_found += 1
            unknown_tokens.append(count[0])
        # Depending on the time of running, we might do the session.commit after each update to avoid
        # blocking the Token table.

    linotp_session.commit()

    print()
    print("{0!s:6} tokens processed.".format(processed))
    print("{0!s:6} counters updated.".format(updated))
    print("{0!s:6} tokens not found.".format(not_found))
    print("List of unknown tokens: {0!s}".format(unknown_tokens))


if __name__ == '__main__':
    main()



"""shorten columns used in UNIQUE constraints in periodictasklastrun, periodictaskoption

Revision ID: d2ae8e54b628
Revises: 1a0710df148b
Create Date: 2018-10-24 11:17:40.279312

"""

# revision identifiers, used by Alembic.
revision = 'd2ae8e54b628'
down_revision = '1a0710df148b'

from alembic import op
import sqlalchemy as sa


def upgrade():
    # First, try to create the table with the shorter field sizes.
    # This will fail if the tables already exist (which will be
    # the case in most installations)
    try:
        op.create_table('periodictasklastrun',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('periodictask_id', sa.Integer(), nullable=True),
        sa.Column('node', sa.Unicode(length=255), nullable=False),
        sa.Column('timestamp', sa.DateTime(), nullable=False),
        sa.ForeignKeyConstraint(['periodictask_id'], ['periodictask.id'], ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('periodictask_id', 'node', name='ptlrix_1'),
        mysql_row_format='DYNAMIC'
        )
        op.create_table('periodictaskoption',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('periodictask_id', sa.Integer(), nullable=True),
        sa.Column('key', sa.Unicode(length=255), nullable=False),
        sa.Column('value', sa.Unicode(length=2000), nullable=True),
        sa.ForeignKeyConstraint(['periodictask_id'], ['periodictask.id'], ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('periodictask_id', 'key', name='ptoix_1'),
        mysql_row_format='DYNAMIC'
        )
        print('Successfully created shortened periodictasklastrun and periodictaskoption.')
    except Exception as exx:
        print('Creation of periodictasklastrun and periodictaskoption with shortened columns failed: {!r}'.format(exx))
        print('This is expected behavior if they were already present.')
    try:
        op.alter_column('periodictasklastrun', 'node',
                   existing_type=sa.VARCHAR(length=256),
                   type_=sa.Unicode(length=255),
                   existing_nullable=False)
        op.alter_column('periodictaskoption', 'key',
                   existing_type=sa.VARCHAR(length=256),
                   type_=sa.Unicode(length=255),
                   existing_nullable=False)
        print('Successfully shortened columns of periodictasklastrun and periodictaskoption.')
    except Exception as exx:
        print('Shortening of periodictasklastrun and periodictaskoption columns failed: {!r}'.format(exx))
        print('This is expected behavior if the columns have already been shorted.')


def downgrade():
    op.alter_column('periodictaskoption', 'key',
               existing_type=sa.Unicode(length=255),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('periodictasklastrun', 'node',
               existing_type=sa.Unicode(length=255),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
